"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

from selenium.webdriver import ChromeOptions
from metatutu.dc.scraping import WebDriverSession

__all__ = ["FullPageCapturer"]

class FullPageCapturer:
    """Capturer of full page."""
    def __init__(self):
        self.init_width = 1280  #initial/exoected width for measurment
        self.init_height = 768  #initial/expected height for measurment
        self.window_width = None  #window width for capturing
        self.window_height = None  #window height for capturing

    def open_session(self, mode):
        """Open a session with container at specific size.

        Default implementation is with Chrome at specific mode.
        Derived classes could override this to meet the needs.

        :param mode: Mode to open session.
            "initial" mode is to open the session for page size measurement.
            "capturing" mode is to open the session for page capturing.
        :returns: Returns a session on success or None on failure.
        """
        session = None
        try:
            #initialize options
            options = ChromeOptions()
            options.add_argument("--incognito")
            if mode == "initial":
                options.add_argument("--window-size={},{}".format(self.init_width, self.init_height))
            elif mode == "capturing":
                options.add_argument("--window-size={},{}".format(self.window_width, self.window_height))
                options.add_argument("--hide-scrollbars")
            else:
                raise Exception()
            options.add_argument("--headless")
            options.add_experimental_option("excludeSwitches", ["enable-logging"])

            #create/open session
            session = WebDriverSession()
            if session.open(options=options) is None: raise Exception()

            #
            return session
        except:
            if session: session.close()
            return None

    def load_page(self, session, url):
        """Load page.

        Default implement is to simply load the page at specific URL.
        Derived classes could override this to add actions to make page
        ready for capture.

        :param session: Session.
        :param url: URL of the page.
        :return: Returns True to continue capturing or False on failure and stop.
        """
        session.get(url)
        return True

    @classmethod
    def capture(cls, url, filepath, init_width=None, init_height=None):
        """Capture a page as image file.

        :param url: URL of the page.
        :param filepath: File path of the output image file.
        :param init_width: Initial/expected width.  This will be used to
            limit the page width.  The actual output will be based on the page
            width gotten under this initial settings for responsive pages.
            If it's None, use default value from class of captuer derived from
            `FullPageCaptuer`.
        :param init_height: Initial/expected height.
        :returns: Returns True on success and False on failure.
        """
        session = None
        try:
            #create capturer
            capturer = cls()
            if init_width: capturer.init_width = init_width
            if init_height: capturer.init_height = init_height

            #pass 1: measure sizes
            session = capturer.open_session("initial")
            if session is None: return False
            result = capturer.load_page(session, url)
            sizes = session.get_sizes()
            print(sizes)
            capturer.window_width = sizes["page.width"] + sizes["window.border.width"]
            capturer.window_height = sizes["page.height"] + sizes["window.border.height"]
            session.close()
            session = None
            if not result: return False

            #pass 2: capture page
            session = capturer.open_session("capturing")
            if session is None: return False
            result = capturer.load_page(session, url)
            if result: result = session.save_screenshot(filepath)
            session.close()
            session = None
            if not result: return False

            #
            return True
        except:
            if session: session.close()
            return False



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

from metatutu.osys import OSUtils, Cmdline
from metatutu.logging import LoggerHelper

class Caller(LoggerHelper):
    """Base class of callers of 3rd party applications."""
    def __init__(self):
        LoggerHelper.__init__(self)

        self.mode = "output"
        self.last_output = {}

    @classmethod
    @property
    def is_windows(cls):
        return OSUtils.is_windows

    @classmethod
    @property
    def is_posix(cls):
        return OSUtils.is_posix

    def run(self, cmdline, run_as_command=False, **kwargs):
        """Run a command with shell or an application without shell.

        :param cmdline: Command line.
        :param run_as_command: Whether to run command line as command or application.
            True to run it as command with shell, False to run it as application
            without shell.
        :returns: Returns return code of the command.  Returns None on failure.
        """
        mode = kwargs.pop("mode", None)
        if mode is None: mode = self.mode
        self.last_output = {}
        return OSUtils.run(cmdline, run_as_command, mode=mode, output=self.last_output, **kwargs)

class ShellCaller(Caller):
    """Caller for OS Shell.

    Since most of basic shell commands have solution to be replaced with Python implementation,
    so this is more like an example on how to create callers with `Caller`.
    """
    def __init__(self):
        Caller.__init__(self)

    def ping(self, target, count=5, size=32):
        """Ping target."""
        cmdline = Cmdline()
        if self.is_windows:
            cmdline.add("ping")
            cmdline.add("-n {}".format(count))
            cmdline.add("-l {}".format(size))
            cmdline.add(target)
        elif self.is_posix:
            cmdline.add("ping")
            cmdline.add("-c {}".format(count))
            cmdline.add("-s {}".format(size - 8))
            cmdline.add(target)
        else:
            return None
        return self.run(cmdline, True)

    def env(self):
        """Get system environment variables."""
        cmdline = Cmdline()
        if self.is_windows:
            cmdline.add("set")
        elif self.is_posix:
            cmdline.add("env")
        else:
            return None
        return self.run(cmdline, True)



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import os
import sys
import getopt
from metatutu.logging import LoggerHelper, Loggers

class CLIApp(LoggerHelper):
    """Base class of CLI applications."""
    def __init__(self):
        LoggerHelper.__init__(self)

        #initialize loggers
        self.loggers = Loggers()
        self.bind_logger(self.loggers)

    def register_logger(self, logger):
        self.loggers.register_logger(logger)

    def init_app(self):
        """Initialize the application.

        This will be called by framework at application starts.
        It could be used to initialize the application environment.

        :returns: Returns True to continue the program, or False to exit.
        """
        return True

    def determine_workflow(self):
        """Determine the workflow to run.

        This will be called by framework after initialization.
        Applicaiton could either use command line or menu to determine
        the workflow to run.

        :returns: Returns the workflow name (str) to run.
            Returns None to exit program.
        """
        return "default"

    def _globals(self):
        """Global namespace for workflow handler searching.

        When dispatching workflow, if the handler is not found within
        the members, it could alternatively search in a global namespace,
        so that the workflow handler could be as a regular function.
        If application needs to have this capability, simple implement
        this with below code:

            def _globals(self): return globals()
        """
        return None

    def run_workflow(self, workflow_name):
        """Run the workflow.

        Framework will call this to run the specific workflow.

        :param workflow_name: Workflow name.
        :returns: Returns program exit code gotten from the workflow.
            It returns None on unexpected error.
        """
        def find_workflow_handler(workflow_name):
            try:
                #handler name
                handler_name = "workflow_" + workflow_name

                #find handler in the member
                if hasattr(self, handler_name):
                    f = getattr(self, handler_name)
                    if callable(f): return f

                #find handler in globals
                x_globals = self._globals()
                if x_globals:
                    f = x_globals.get(handler_name)
                    if callable(f): return f
            except:
                pass
            return None

        try:
            f = find_workflow_handler(workflow_name)
            if f is None: return self.on_workflow_not_dispatched(workflow_name)
            return f()
        except:
            return 1

    def workflow_default(self):
        """Handler of default workflow.

        "default" is the workflow name of this framework.
        So for an application with single workflow, simple implement this function
        could make it happens.

        :returns: Returns the exit code of the program.
        """
        return 0

    def on_workflow_not_dispatched(self, workflow_name):
        """Handler for case that workflow is not dispatched.

        This will be called by framework when workflow could not be dispatched.

        :param workflow_name: Workflow name.
        :returns: Returns the exit code of the program.
        """
        return 1

    def cleanup_app(self):
        """Clean up the application.

        This will be called by framework before exiting program.
        """
        #cleanup loggers
        self.loggers.close()

    def main(self):
        """Application framework."""
        exit_code = 1
        try:
            #initialize application
            if not self.init_app(): raise Exception()

            #determine workflow
            workflow_name = self.determine_workflow()
            if workflow_name is None: raise Exception()

            #run workflow
            exit_code = self.run_workflow(workflow_name)
        except:
            pass
        finally:
            #cleanup application
            self.cleanup_app()
        os._exit(exit_code)

    @classmethod
    def parse_command_line(cls, parts):
        """Parse command line.

        Command line syntax is as:
        program <command 1> <command 2> ... [option 1] [option 2] ...

        Commands are required and options are optional.  Commands are
        always before options.

        :param parts: This is a list of argument definitions.
            Each list item specify one arguments to be read from command line.
            The item is in a tuple as (key, short, long).
            "key" is used to query the argument value.
            "__" prefixed key names are reserved for internal use.
            "short" is the short format of option.  eg. "-t:" expects an option
            given as "-t <value>", "-i" expects an option wihtout value.
            "long" is the long format of option.  eg. "time=" expects an option
            given as "--time=<value>", "id" expects an option given as "--id".
            For more information of "short" and "long", check `getopt.getopt()`.
            When "short" and "long" are both empty (""), it specifies a command.
        :returns: Returns a dict as {key: value}.
            "key" is what specified in `parts` for commands and options.
            The value is typically a str value read from command line.
            If the option is not with value, with its existence in the dict,
            it means it had been specified in the command line.
            There are some special keys: "__cmd__" is with the full command line,
            "__argn__" is with the number of arguments and "__argv0__" is
            the first item of `sys.argv[]` which is the program.  "__args__" is
            with remaining arguments not parsed.
            If there is no argument at all (run program directly), it will returns
            "__cmd__", "__argn__" and "__argv0__" only.
            If there is argument given but not matching the syntax as expected,
            it will return None.  If there is other unknown error, it will return
            None as well.
        """
        try:
            #process parts
            parts_commands = []
            parts_options = []
            shortopts = ""
            longopts = []
            for name, short, long in parts:
                if short == "" and long == "":
                    parts_commands.append(name)
                else:
                    shortopts += short
                    longopts.append(long)
                    formats = []
                    if short != "": formats.append(short.replace(":", ""))
                    if long != "": formats.append("--" + long.replace("=", ""))
                    parts_options.append((name, formats))

            #get basic information
            r = {}
            r["__cmd__"] = " ".join(sys.argv)
            r["__argn__"] = len(sys.argv)
            r["__argv0__"] = sys.argv[0]
            if len(sys.argv) <= 1: return r

            #get commands
            command_count = len(parts_commands)
            if len(sys.argv) < 1 + command_count: return None
            for i in range(0, command_count):
                r[parts_commands[i]] = sys.argv[1 + i]

            #get options
            opts, args = getopt.getopt(sys.argv[1 + command_count:], shortopts, longopts)
            for opt_name, opt_value in opts:
                for name, formats in parts_options:
                    if opt_name in formats:
                        r[name] = opt_value

            #get args
            r["__args__"] = args

            #
            return r
        except:
            return None



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""


"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import matplotlib
import matplotlib.pyplot as plt

__all__ = ["Figure"]

class Figure:
	"""Base class of figure classes.

	Figure is basically a canvas to be with one or multiple charts on it
	to visualize one set of data, so that reader is able to focus analysing
	the data from different angle.

	This class is wrapping up the common behaviors of matplotlib usage and
	defined a simple framework to help data analyzer to focus on creating
	chart/figure rather than knowing details on how to use matplotlib.  It
	also works around some existing issue in matplotlib in subplots.

	A typical usage of Figure class is as below:

	* Create a class deriving from ``class Figure``
	* Override ``on_paint()`` to include the logic to draw charts
	* In ``on_paint()``, call ``add_chart()`` to get the ``ax`` object
	* Use ``ax`` object to draw specific chart with matplotlib

	.. hint::
		Override ``_get_chart_rect()`` could define the customized layout
		not necessarys to be in cells with rows and columns.

	:ivar figsize: Size of the figure in inches.
		It specifies the size in a tuple as (width, height).
	:ivar layout: Layout of the figure.
		It's similar to subplots in matplotlib, defining number of charts
		by (rows, cols).
	:ivar no_show: Defining whether the figure will be displayed.
		It's useful on Jupyter as the figure will be displayed immediately.
		When some figure is not needed to be displayed, for example, to be
		created and further processed, it could set this option to False to
		prevent it be displayed.
	:ivar filepath: File path.  If it is specified, the figure will be saved
		as file.  If it is None, then file saving action will be skipped.
	:ivar save_dpi: DPI for file saving.
	:ivar save_facecolor: Face color of figure when it is saved as file.
	:ivar save_tight: Whether the figure will be cut as "tight" model when
		it's saved to file.
	:ivar data_binding: This variable is used to bundle the source data
		of chart(s) on figure as part of framework defined.
	"""

	def __init__(self):
		# config
		self.figsize = (20, 10)
		self.layout = (1, 1)
		self.no_show = False
		self.filepath = None
		self.save_dpi = 72
		self.save_facecolor = "white"
		self.save_tight = False
		self.data_binding = None

		# control
		self._f = None

	def _get_chart_rect(self, pos=None):
		if pos is None:
			return (0.0, 0.0, 1.0, 1.0)
		elif type(pos) is tuple:
			rows, cols = self.layout
			row, col = pos
			if row >= 0 and row < rows and col >= 0 and col < cols:
				l = col / cols
				t = row / rows
				r = (col + 1) / cols
				b = (row + 1) / rows
				return (l, t, r, b)
		else:
			rows, cols = self.layout
			row = int((pos - 1) / cols)
			col = (pos - 1) % cols
			return self._get_chart_rect((row, col))
		return None

	def add_chart(self, title=None, pos=None, padding=(0.125, 0.12, 0.10, 0.11)):
		"""Add a chart.

		It's basically to define an axes system for drawing on the figure (canvas).

		:param title: Chart title.  If it's given, chart title will be created.
		:param pos: Position of the chart.
			If it's None, then the full figure (canvas) will be used for the chart.
			If it's given as a number (int), it will be treated as an index
			value (starting from 1), based on the figure layout, it will locate
			the right cell on figure.  (left to right, up to down)
			If it's given as (row, col), it will locate to the cell on figure
			based on the figure layout.
		:param padding: Padding of the chart area.
			It specifies the padding on each side as (left, top, right, bottom)
			in percentage.  eg. (0.1, 0.1, 0.1, 0.1) means 10% of width padding
			will be on left and right sides, and 10% of height padding will be
			on top and bottom sides.
		"""
		rect = self._get_chart_rect(pos)
		if rect is None: return None
		l, t, r, b = rect
		w = r - l
		h = b - t
		ax = self._f.add_axes([
			l + w * padding[0],
			1 - (b - h * padding[3]),
			w * (1.0 - padding[0] - padding[2]),
			h * (1.0 - padding[1] - padding[3])])
		if title: plt.title(title)
		return ax

	def save_figure(self, filepath):
		"""Save figure as file.

		:param filepath: File path.
		"""
		if filepath:
			self._f.savefig(
				filepath,
				dpi=self.save_dpi,
				facecolor=self.save_facecolor,
				bbox_inches=("tight" if self.save_tight else None))

	def on_create_figure(self):
		"""Event handler called when a figure is being created."""
		self._f = plt.figure(figsize=self.figsize)

	def on_paint(self):
		"""Event handler called when a figure is created and needs to be painted.

		:returns: Returns False to stop further workflow.
		"""
		pass

	def on_save_figure(self):
		"""Event handler called when a figure is being saved to file after painted."""
		self.save_figure(self.filepath)

	def on_show_figure(self):
		"""Event handler called when a figure is being displayed after painted."""
		if not self.no_show: self._f.show(warn=False)

	def paint(self, show_error=False):
		"""Paint figure.

		Call this to run the workflow defined by framework to create figure,
		paint charts, save file and display.

		:param show_error: Whether to display errors (typically from matplotlib).
		"""
		old_backend = matplotlib.get_backend()
		backend_switched = False
		try:
			# switch backend if needed
			if self.no_show:
				matplotlib.use("agg")
				backend_switched = True

			# create figure
			self.on_create_figure()

			# paint
			painted = True
			r = self.on_paint()
			if r is not None:
				if r == False: painted = False
			if painted:
				# save figure
				self.on_save_figure()

				# show figure
				self.on_show_figure()

			# delete figure
			del self._f
			self._f = None
		except Exception as e:
			if show_error: print(e)
		finally:
			if backend_switched:
				matplotlib.use(old_backend)





"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

__NAME__ = "kimi"
__DESC__ = "metatutu library"
__VERSION__ = "0.0.5"
__LICENSE__= "MIT License"
__AUTHOR__ = "Wooloo Studio"
__AUTHOR_EMAIL__ = "max.wu@wooloostudio.com"
__URL__ = "https://github.com/metatutu/metatutu"
__PROJECT_URLS__ = {
	"Documentation": "https://github.com/metatutu/metatutu",
	"Source Code": "https://github.com/metatutu/metatutu",
	"Bug Tracker": "https://github.com/metatutu/metatutu/issues",
	"Forum": "https://github.com/metatutu/metatutu/issues",
	"Donate": "https://github.com/metatutu/metatutu"
}
__BUILD__ = 78


"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import os

__all__ = ["about"]

def about():
	about = {}
	here = os.path.abspath(os.path.dirname(__file__))
	filepath = os.path.join(here, "__version__.py")
	with open(filepath, "r", encoding="utf-8") as f:
		exec(f.read(), None, about)
	print(about["__NAME__"])
	print("Version: " + about["__VERSION__"])
	print("Author: " + about["__AUTHOR__"])
	print("Email: " + about["__AUTHOR_EMAIL__"])



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import os
import threading
import time
import win32api
import win32con
from metatutu.app.cli import CLIApp
from metatutu.logging import ConsoleLogger

class Servicelet(CLIApp):
    """Base class of servicelet applications.

    Servicelet is a console based application which could be used to run some
    background service without user interaction.  The console could display
    some service status information, and the information could be logged as well
    in other forms.

    This class defines the framework of a servicelet application, so that derived
    classes could focus on the service logic.  Derviced class need to override
    `service_processor()` to include the service logic.

    There are 2 ways of user termination defined:
    1. User press CTRL+C in the active console.

    2. User press CTRL+ALT+Z anywhere in the operating system (support Windows
    only for now).  This is optional to the servicelet, and it could be turn
    off by set `self.listen_global_stop_request` to False.

    Both will set `stop_requested` flag.  So in the implementation of
    `service_processor()`, it needs to check the `self.stop_requested` time to
    time and exit the process smoothly.

    Servicelet information could be set in `self.service_info`.  `name` is
    mandatory and others are optional though we strongly suggest to have them set.
    """

    def __init__(self):
        CLIApp.__init__(self)

        self.service_info = {
            "name": None,
            "version": None,
            "copyright": None,
            "license": None
        }

        self.listen_global_stop_request = True

        console_logger = ConsoleLogger()
        console_logger.include_timestamp = False
        self.register_logger(console_logger)

        self._control_lock = threading.Lock()
        self._stop_requested = False

    @property
    def stop_requested(self):
        with self._control_lock:
            return self._stop_requested

    def service_processor(self):
        """Service processor.

        This will be called by framework when starting the service program.
        It will be run in thread.  Derived class must override this to implement the
        process logic of service application.

        The process logic could be a single process taking long time, or repeating or
        even in dead loop.  The process needs to check stop request time to time, and
        exit the process smoothly.  The main thread will take care of the terminal
        signals from users and raise the stop request accordingly.

        When this process routine returns, the main thread and program will
        exit as well.
        """
        raise NotImplementedError

    def __service_processor(self):
        self.log("Entering service processr...")
        try:
            self.service_processor()
        except Exception as ex:
            self.exception(ex)
        self.log("Leaving service processor...")

    def workflow_default(self):
        try:
            #service info
            si = self.service_info
            if si.get("name") is None: return 1
            self.log("Service Name: {}".format(si.get("name")))

            if si.get("version"): self.log("Version: {}".format(si.get("version")))
            if si.get("copyright"): self.log(si.get("copyright"))
            if si.get("license"): self.log(si.get("license"))

            #start service process
            self.log("Starting service processor...")
            self.log("Preparing...", 1)
            self._stop_requested = False
            th = threading.Thread(target=self.__service_processor)
            self.log("Starting...", 1)
            th.start()

            #wait for service stopped or terminated
            while True:
                try:
                    #check global terminate signal (Windows Only)
                    if self.listen_global_stop_request and os.name == "nt":
                        key_ctrl_down = (win32api.GetKeyState(win32con.VK_CONTROL) & 0x80 != 0)
                        key_alt_down = (win32api.GetKeyState(win32con.VK_MENU) & 0x80 != 0)
                        key_z_down = (win32api.GetKeyState(90) & 0x80 != 0)
                        if key_ctrl_down and key_alt_down and key_z_down:
                            self.info("CTRL+ALT+Z is pressed.  Stopping the service processor...")
                            with self._control_lock: self._stop_requested = True
                            break

                    #check processor state
                    if not th.is_alive(): break

                    #sleep
                    time.sleep(0.1)
                except KeyboardInterrupt:
                    self.info("CTRL+C is pressed.  Stopping the service processor...")
                    with self._control_lock: self._stop_requested = True
                    break

            #stop service processor
            th.join()
            self.log("Service processor is stopped.")

            #
            return 0
        except Exception as ex:
            self.exception(ex)
            return 1



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import time

class Clocker:
	"""Clocker for debugging."""

	def __init__(self):
		self.checkpoints = []
		self.reset()

	def reset(self):
		"""Reset."""
		self.checkpoints.clear()
		self.record()

	def record(self, label=""):
		"""Record a checkpoint.

		:param label: Label of the checkpoint to identify the phase
			from previous checkpoint to current one.
		"""
		self.checkpoints.append((label, time.perf_counter()))

	def __enter__(self):
		self.reset()
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.record()

	def results(self):
		"""Get summary of timing results.

		:returns: A list of timing results for each phase (end by checkpoint).
			Each result is in format of (label, duration in seconds).
		"""
		results = []
		if len(self.checkpoints) <= 1: return results
		last_timestamp = None
		for (label, timestamp) in self.checkpoints:
			if last_timestamp is not None:
				results.append((label, timestamp - last_timestamp))
			last_timestamp = timestamp
		return results

	def results_text(self):
		"""Get summary of timing results in text.

		:returns: Text of summary of timing results for display.
		"""
		results = self.results()
		text = ""
		index = 1
		for (label, duration) in results:
			if text != "": text += "\r\n"
			if label == "":
				text += "(#{}): {:.1f} seconds".format(index, duration)
			else:
				text += "{}: {:.1f} seconds".format(label, duration)
			index += 1
		return text

	def duration(self, label=None):
		"""Get duration of the phase.

		:param label: Checkpoint label of the phase to be queried.
			If it's None, return the duration of the last phase.

		:returns: The duration of the phase in seconds.
			Returns None when the checkpoint is not found.
		"""
		results = self.results()
		results.reverse()
		if label is None:
			if len(results) > 0: return results[0][1]
		else:
			for (label2, duration) in results:
				if label2 == label: return duration
		return None



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import smtplib
from email.utils import formataddr, formatdate
from email.mime.multipart import MIMEMultipart
from email.header import Header
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from metatutu.logging import LoggerHelper, MessageLevel
from metatutu.fsds import FileSystemUtils

__all__ = ["Message", "Sender"]

class Message:
    """E-mail message."""
    def __init__(self):
        self.sender = (None, "")
        self.to = []
        self.cc = []
        self.bcc = []
        self.subject = ""
        self.attachments = []
        self.body = ("plain", "")
        self.timeval = None
        self.localtime = False
        self.usegmt = False
        self.importance = None
        self.sensitivity = None
        self.delivery_notification = False
        self.read_notification = False

    def set_sender(self, address, name=None):
        """Set sender.

        :param address: E-mail address of the sender.
        :param name: Name of the sender.
        """
        self.sender = (name, address)

    def add_receiver(self, address, name=None, type="to"):
        """Add a receiver.

        :param address: E-mail address of the receiver.
        :param name: Name of the receiver.
        :param type: Type of receiver.  Valid options are "to", "cc" and "bcc".
        """
        if type == "to":
            self.to.append((name, address))
        elif type == "cc":
            self.cc.append((name, address))
        elif type == "bcc":
            self.bcc.append((name, address))

    def add_receivers(self, receivers, type="to"):
        """Add receivers.

        :param receivers: List of receivers with items as (name, address).
        :param type: Type of receivers.  Valid options are "to", "cc" and "bcc".
        """
        if type == "to":
            self.to += receivers
        elif type == "cc":
            self.cc += receivers
        elif type == "bcc":
            self.bcc += receivers

    def set_subject(self, subject):
        """Set subject.

        :param subject: E-mail subject.
        """
        self.subject = subject

    def add_attachment(self, filepath, as_filename):
        """Add an attachment.

        :param filepath: The path of the file to be attached.
        :param as_filename: The filename to be displayed in the e-mail message.
        :returns: Returns True on success and False on failure.
        """
        data = FileSystemUtils.load_file_bytes(filepath)
        if data is None: return False
        self.attachments.append((data, as_filename))
        return True

    def set_body(self, body, type="plain"):
        """Set E-mail body.

        :param body: E-mail body.
        :param type: Body type.  Valid options are "plain" and "html".
        """
        if type == "plain":
            self.body = ("plain", body)
        elif type == "html":
            self.body = ("html", body)

    def set_date(self, timeval=None, localtime=False, usegmt=False):
        """Set date.

        :param timeval: Timestamp.  If it's None, use current time.
        :param localtime: Whether timeval is in local time or UTC.
        :param usegmt: Whether timezone is written out as ascii string.
        """
        self.timeval = timeval
        self.localtime = localtime
        self.usegmt = usegmt

    def set_importance(self, importance):
        """Set importance.

        :param importance: Valid options are None, "high" and "low".
        """
        self.importance = importance

    def set_sensitivity(self, sensitivity):
        """Set sensitivity.

        :param sensitivity: Valid options are None, "personal", "private" and "confidential".
        """
        self.sensitivity = sensitivity

    def get_sender_address(self):
        """Get sender's E-mail address."""
        if self.sender:
            return self.sender[1]
        else:
            return ""

    def get_receiver_addresses(self):
        """Get list of receivers' e-mail addresses."""
        addresses = []
        for item in self.to:
            address = item[1]
            if address not in addresses: addresses.append(address)
        for item in self.cc:
            address = item[1]
            if address not in addresses: addresses.append(address)
        for item in self.bcc:
            address = item[1]
            if address not in addresses: addresses.append(address)
        return addresses

    def get_formatted_message(self):
        """Create a formatted message object."""
        def format_address(item):
            name, address = item
            if name:
                return formataddr((Header(name, "utf-8").encode(), address))
            else:
                return address

        def format_addresses(items):
            addresses = []
            for item in items:
                addresses.append(format_address(item))
            return ",".join(addresses)

        # body
        body_type, body_content = self.body
        if body_type == "html":
            part_body = MIMEText(body_content, "html", "utf-8")
        else:
            part_body = MIMEText(body_content, "plain", "utf-8")

        if len(self.attachments) > 0:
            message = MIMEMultipart()
            message.attach(part_body)
        else:
            message = part_body

        # date
        message["Date"] = formatdate(self.timeval, self.localtime, self.usegmt)

        # sender
        message["From"] = format_address(self.sender)

        # receivers
        message["To"] = format_addresses(self.to)
        message["Cc"] = format_addresses(self.cc)

        # subject
        message["Subject"] = Header(self.subject, "utf-8")

        # importance
        if self.importance:
            importance = self.importance.lower()
            if importance == "high":
                message["X-Priority"] = "1"
                message["Importance"] = "high"
            elif importance == "low":
                message["X-Priority"] = "5"
                message["Importance"] = "low"

        # sensitivity
        if self.sensitivity:
            sensitivity = self.sensitivity.lower()
            if sensitivity == "personal":
                message["Sensitivity"] = "Personal"
            elif sensitivity == "private":
                message["Sensitivity"] = "Private"
            elif sensitivity == "confidential":
                message["Sensitivity"] = "Company-Confidential"

        # delivery notification
        if self.delivery_notification:
            message["Return-Receipt-To"] = format_address(self.sender)

        # read notification
        if self.read_notification:
            message["Disposition-Notification-To"] = format_address(self.sender)

        # attachments
        for attachment in self.attachments:
            att = MIMEApplication(attachment[0])
            att["Content-Type"] = "application/octet-stream"
            att.add_header("Content-Disposition", "attachment", filename=attachment[1])
            message.attach(att)

        #
        return message

class Sender(LoggerHelper):
    """Message sender."""
    def __init__(self):
        LoggerHelper.__init__(self)

        # config
        self.smtp_host = None
        self.smtp_ssl = True
        self.smtp_port = None
        self.smtp_user = None
        self.smtp_password = None

        # control
        self._client = None

    def __del__(self):
        self.disconnect()

    def connect(self):
        """Connect to SMTP server and login.

        :returns: Returns SMTP client handler on success and None on failure.
        """
        try:
            if self._client: return self._client

            # connect SMTP server
            self.debug("Connecting SMTP server...")
            if self.smtp_ssl:
                client = smtplib.SMTP_SSL(self.smtp_host, self.smtp_port)
            else:
                client = smtplib.SMTP(self.smtp_host, self.smtp_port)

            # login
            self.debug("Login to SMTP server...")
            result = client.login(self.smtp_user, self.smtp_password)
            print(result)
            if result is None:
                self.error("Login failed!", level=MessageLevel.DEBUG)
                client.quit()
                return None
            if result[0] != 235:
                self.error("Login failed! ({})".format(result[0]), level=MessageLevel.DEBUG)
                client.quit()
                return None
            self._client = client
            return self._client
        except Exception as ex:
            self.exception(ex)
            return None

    def disconnect(self):
        """Disconnect to SMTP server."""
        try:
            if self._client is None: return

            # disconnect SMTP server
            self.debug("Disconnecting SMTP server...")
            self._client.quit()
            self._client = None
        except:
            self._client = None

    @property
    def client(self):
        return self.connect()

    def send(self, message, reset_connection=False):
        """Send message.

        :param message: Message object.
        :param reset_connection: Whether to reset the connection.
        :returns: Returns True on success and False on failure.
        """
        try:
            # get client
            if reset_connection: self.disconnect()
            client = self.client
            if client is None: return False

            # send
            print("sending")
            client.sendmail(
                message.get_sender_address(),
                message.get_receiver_addresses(),
                message.get_formatted_message().as_string()
            )
            print("sent")

            #
            return True
        except Exception as ex:
            self.exception(ex)
            return False



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

from abc import ABC, abstractclassmethod
import os
import time
import requests
import hashlib
import chardet
import json
from selenium import webdriver
from selenium.webdriver.common.by import By
from bs4 import BeautifulSoup
from metatutu.logging import LoggerHelper, MessageLevel
from metatutu.fsds import FileSystemUtils, FileSystemDataStore, TempFile
from metatutu.images import ImageUtils

__all__ = [
    "Cache", "FileSystemCache",
    "Session", "HttpSession", "WebDriverSession",
]

class Cache(ABC):
    """Base class of cache classes.

    Cache is a component in scraping framework to store the raw content obtained
    from web during scraping process, so that it could be reused next time.
    For example, scraping application could cache all pages (aka. source HTML) had
    be load, so when the process stops for some reason, rerun the process doesn't
    need to load the pages already cached.

    Derived class needs to implement below interfaces with specific persistence
    strategy:

    * ``set()``: Set key value.
    * ``remove()``: Remove the key value.
    * ``get()``: Get key value.
    * ``clear()``: Clear all/expired key values..
    """
    @abstractclassmethod
    def set(self, key, value, text_format=False):
        """Set key value.

        :param key: Key.
        :param value: Value.
        :param text_format: Whether the value is in string or bytes format.
        """
        pass

    @abstractclassmethod
    def remove(self, key):
        """Remove the key value.

        :param key: Key.
        """
        pass

    @abstractclassmethod
    def get(self, key, valid_period=0, text_format=False):
        """Get key value.

        :param key: Key.
        :param valid_period: Valid period of cached value (seconds).
            If it's 0, the key value is valid forever.
        :param text_format: Whether the value is in string or bytes format.

        :return: It returns the cached value if it's hit.  Otherwise, it returns None.
        """
        pass

    @abstractclassmethod
    def clear(self, valid_period=0):
        """Clear all/expired key values.

        :param valid_period: Valid period of cached value (seconds).
            If it's 0, all the key values will be removed.
        """
        pass

    @classmethod
    def make_key(cls, resource_id):
        """Make a unique key of a resource.

        :param resource_id: This is a string to identify the resource.
            Typically, it is a URL of a page in web scraping application.
        """
        return hashlib.md5(resource_id.encode(encoding="utf-8")).hexdigest()

class FileSystemCache(Cache):
    """Cache with persistence on file system.

    All cached data will be stored in a folder on file system.
    Each key value will be saved as a single file with the name as same as the key.

    :ivar root_folderpath: Root folder path.
    """
    def __init__(self):
        # config
        self.root_folderpath = None

    def _get_filepath(self, key):
        fsds = FileSystemDataStore(self.root_folderpath)
        return fsds.get_path(key, True)

    def set(self, key, value, text_format=False):
        fpath = self._get_filepath(key)
        if value is None:
            FileSystemUtils.delete_file(fpath)
        else:
            if text_format:
                FileSystemUtils.save_file_contents(fpath, value)
            else:
                FileSystemUtils.save_file_bytes(fpath, value)

    def remove(self, key):
        fpath = self._get_filepath(key)
        FileSystemUtils.delete_file(fpath)

    def get(self, key, valid_period=0, text_format=False):
        fpath = self._get_filepath(key)
        if not FileSystemUtils.file_exists(fpath): return None
        if valid_period > 0:
            if (time.time() - os.path.getctime(fpath)) > valid_period: return None
        if text_format:
            return FileSystemUtils.load_file_contents(fpath)
        else:
            return FileSystemUtils.load_file_bytes(fpath)

    def clear(self, valid_period=0):
        fpaths = FileSystemUtils.list_files(self.root_folderpath)
        t_now = time.time()
        if fpaths is None: return
        if valid_period > 0:
            for fpath in fpaths:
                if (t_now - os.path.getctime(fpath)) > valid_period:
                    FileSystemUtils.delete_file(fpath)
        else:
            for fpath in fpaths:
                FileSystemUtils.delete_file(fpath)

class Session(ABC, LoggerHelper):
    """Base class of session classes.

    In this scrapping framework, session is defined as the controller to access
    the web.  It's used from the first access to the specific website till the
    last access.  One session instance is with one base method.  Method could
    be implemented with HTTP requests, seleiumn webdrivers, and so on.
    """
    def __init__(self):
        LoggerHelper.__init__(self)

        # control
        self._handle = None

    def __del__(self):
        self.close()

    @abstractclassmethod
    def open(self, **kwargs):
        """Open the session.

        :returns: Returns the open session handle on success or None on failure.
        """
        return None

    @abstractclassmethod
    def close(self):
        """Close the session."""
        pass

    @abstractclassmethod
    def _get_page(self, url, cache, format, **kwargs):
        return None

    def get_page(self, url, cache=None, format="soup", **kwargs):
        """Get page content.

        :param url: URL.
        :param cache: Page cache.
        :param format: Return format of page content.
            Valid options are "html" and "soup".
        :param kwargs: Other parameters.
        :returns: Returns page content in specified format or None on failure.
        """
        if format == "html":
            return self._get_page(url, cache, "html", **kwargs)
        elif format == "soup":
            try:
                html = self.get_page(url, cache, "html", **kwargs)
                if html is None: return None
                return BeautifulSoup(html, "html.parser")
            except Exception as ex:
                self.exception(ex)
                return None
        else:
            return self._get_page(url, cache, format, **kwargs)

    @property
    def handle(self):
        return self.open()

    @classmethod
    def wait(cls, secs=1.0):
        """Wait.

        :param secs: Time to wait, in seconds.
        """
        time.sleep(secs)

    @classmethod
    def bytes_str(cls, content_bytes, encoding=None):
        """Convert bytes to str.

        :param content_bytes: Content in bytes.
        :param encoding: Encoding of content.
            If it's None, it will detect the encoding automatically.
        :returns: The content in string.
        """
        try:
            if content_bytes is None: return ""
            if encoding:
                if encoding == "ISO-8859-1": encoding = None
            if encoding is None: encoding = chardet.detect(content_bytes)["encoding"]
            if encoding == "GB2312": encoding = "GBK"
            content_str = str(content_bytes, encoding, errors='replace')
        except:
            content_str = str(content_bytes, errors='replace')
        return content_str
    
    @classmethod
    def json_loads(cls, text):
        try:
            return json.loads(text)
        except:
            return None

class HttpSession(Session):
    """Session with HTTP requests."""
    def __init__(self):
        Session.__init__(self)

        # config
        self.user_agent = None
        self.ok_status_codes = [200]

    def open(self, **kwargs):
        try:
            if self._handle: return self._handle
            self.debug("Opening HTTP session...")
            self._handle = requests.Session()
            return self._handle
        except Exception as ex:
            self.exception(ex)
            self._handle = None
            return None

    def close(self):
        if self._handle:
            self.debug("Closing HTTP session...")
            self._handle = None

    def response_is_ok(self, response):
        if response:
            return response.status_code in self.ok_status_codes
        else:
            return False

    def request(self, method, url, **kwargs):
        """HTTP request.

        :param method: The method of HTTP request.
            Valid options are "GET" and "POST".
        :param url: URL to be requested.
        :param retries: Number of retries when request was failed.
            If it's None, request will not be retried.
        :param retry_wait: Time to wait between retries.
            If it's None, no wait will happen.
        :param kwargs: Other request parameters.
        :returns: Response object from HTTP request.
            Returns None for invalid calls.
        """
        # check method
        if method not in ["GET", "POST"]: return None

        # check mode
        if "retries" in kwargs:
            # request with retries
            retries = kwargs.pop("retries", 0)
            if retries < 0: return None

            retry_wait = kwargs.pop("retry_wait", 0)

            response = self.request(method, url, **kwargs)
            retry_count = 0
            while True:
                if response:
                    if response.status_code in self.ok_status_codes: break
                if retry_count >= retries: return None
                retry_count += 1
                if retry_wait > 0: self.wait(retry_wait)
                kwargs["call_tag"] = "retry {}/{}".format(retry_count, retries)
                response = self.request(method, url, **kwargs)
            return response
        else:
            # request without retry
            try:
                if url is None: return None
                if url == "": return None

                s = self.handle
                if self.user_agent: s.headers["User-Agent"] = self.user_agent

                call_tag = kwargs.pop("call_tag", None)
                if call_tag:
                    self.debug("HTTP {} ({}): {}".format(method, call_tag, url))
                else:
                    self.debug("HTTP {}: {}".format(method, url))

                if method == "GET":
                    response = s.get(
                        url,
                        **kwargs)
                elif method == "POST":
                    response = s.post(
                        url,
                        data=kwargs.get("data", None),
                        json=kwargs.get("json", None),
                        **kwargs)
                else:
                    response = None
                self.debug("method={}\nurl={}\nstatus={}".format(method, url, response.status_code), 1)

                return response
            except Exception as ex:
                self.exception(ex)
                return None

    def get(self, url, retries=0, **kwargs):
        """HTTP GET request."""
        kwargs["retries"] = retries
        return self.request("GET", url, **kwargs)

    def post(self, url, data=None, json=None, retries=0, **kwargs):
        """HTTP POST request."""
        kwargs["data"] = data
        kwargs["json"] = json
        kwargs["retries"] = retries
        return self.request("POST", url, **kwargs)

    def _get_page(self, url, cache, format, **kwargs):
        if format == "html":
            # read from cache
            if cache:
                key = Cache.make_key(url)
                content = cache.get(key)
                if content: return self.bytes_str(content)

            # read from online
            retries = kwargs.pop("retries", 0)
            response = self.get(url, retries, **kwargs)
            if response:
                if response.status_code == 200:
                    if cache:
                        key = Cache.make_key(url)
                        cache.set(key, response.content)
                    return self.bytes_str(response.content, response.encoding)

        #
        return None

    @classmethod
    def quick_download(cls, url, filepath, retries=0, create_parent=True, **kwargs):
        """Download an object and save it to a file."""
        success = False
        try:
            session = HttpSession()
            response = session.get(url, retries, **kwargs)
            if response:
                if response.status_code == 200:
                    success = FileSystemUtils.save_file_bytes(filepath, response.content, create_parent)
            session.close()
        except:
            pass
        return success

class WebDriverSession(Session):
    """Session with selenium web driver."""
    def __init__(self):
        Session.__init__(self)

        # config
        self.webdriver = "chrome"
        self.default_timeout = 10.0
        self.default_wait = 0.5
        self.default_by = By.XPATH

    def open(self, **kwargs):
        try:
            if self._handle: return self._handle
            self.debug("Opening web driver ({}) session...".format(self.webdriver))
            if self.webdriver == "chrome":
                self._handle = webdriver.Chrome(**kwargs)
            elif self.webdriver == "firefox":
                self._handle = webdriver.Firefox(**kwargs)
            elif self.webdriver == "edge":
                self._handle = webdriver.Edge(**kwargs)
            else:
                raise Exception("Invalid webdriver.")
            return self._handle
        except Exception as ex:
            self.exception(ex)
            return None

    def close(self):
        try:
            if self._handle is None: return
            self.debug("Closing web driver ({}) session...".format(self.webdriver))
            self._handle.close()
            self._handle = None
        except Exception as ex:
            self.exception(ex)
            self._handle = None

    def get(self, url):
        """Open URL."""
        try:
            self.debug("Web Driver ({}) GET: {}".format(self.webdriver, url))
            self.handle.get(url)
            return True
        except Exception as ex:
            self.exception(ex)
            return False

    def get_html(self):
        return self.execute("return document.documentElement.outerHTML")

    def get_soup(self):
        try:
            html = self.get_html()
            if html is None: return None
            return BeautifulSoup(html, "html.parser")
        except Exception as ex:
            self.exception(ex)
            return None

    def _get_page(self, url, cache, format, **kwargs):
        if format == "html":
            # read from cache
            if cache:
                key = Cache.make_key(url)
                content = cache.get(key, text_format=True)
                if content: return content

            # read from online
            wait = kwargs.pop("wait", 0)
            self.get(url)
            self.wait(wait)
            html = self.get_html()
            if html:
                if cache:
                    key = Cache.make_key(url)
                    cache.set(key, html, text_format=True)
                return html

        #
        return None

    def get_url(self, leading_wait=2.0, timeout=None, wait=None):
        """Get current URL.

        :param leading_wait: Wait time before getting URL, in seconds.
        :param timeout: Timeout, in seconds.
        :param wait: Wait time between each retry, in seconds.

        :returns: Returns current URL or None on failure.
        """
        if timeout is None: timeout = self.default_timeout
        if wait is None: wait = self.default_wait
        if leading_wait > 0.0: time.sleep(leading_wait)
        ts_0 = time.perf_counter()
        url = None
        while True:
            try:
                url = self.handle.current_url
                if url:
                    self.debug("Current URL is: {}".format(url))
                    return url
            except:
                pass

            if time.perf_counter() - ts_0 > timeout:
                self.warning("URL is not obtained!", level=MessageLevel.DEBUG)
                return None

            time.sleep(wait)

    def find_element(self, value, base=None, timeout=None, wait=None, by=None):
        """Find element from base node.

        :param value: Element to ne searched.
        :param base: Base node.  If it's None, search from root.
        :param timeout: Timeout, in seconds.  It will keep trying to find the
            expected element until timeout.
        :param wait: Wait time between each retry, in seconds.
        :param by: Value type.

        :returns: Returns the element found or None.
        """
        if timeout is None: timeout = self.default_timeout
        if wait is None: wait = self.default_wait
        if by is None: by = self.default_by
        ts_0 = time.perf_counter()
        if base is None: base = self.handle
        while True:
            try:
                element = base.find_element(by, value)
                if element: return element
            except:
                pass

            if time.perf_counter() - ts_0 > timeout:
                self.warning("Element '{}' is not found!".format(value), level=MessageLevel.DEBUG)
                return None

            time.sleep(wait)

    def find_elements(self, value, base=None, timeout=None, wait=None, by=None):
        """Find elements from base node.

        :param value: Elements to ne searched.
        :param base: Base node.  If it's None, search from root.
        :param timeout: Timeout, in seconds.  It will keep trying to find the
            expected element until timeout.
        :param wait: Wait time between each retry, in seconds.
        :param by: Value type.

        :returns: Returns the elements found or None.
        """
        if timeout is None: timeout = self.default_timeout
        if wait is None: wait = self.default_wait
        if by is None: by = self.default_by
        ts_0 = time.perf_counter()
        if base is None: base = self.handle
        while True:
            try:
                elements = base.find_elements(by, value)
                if elements:
                    self.debug("{} elements had been found as '{}'.".format(len(elements), value))
                    return elements
            except:
                pass

            if time.perf_counter() - ts_0 > timeout:
                self.warning("Element(s) '{}' is not found.".format(value), level=MessageLevel.DEBUG)
                return None

            time.sleep(wait)

    def execute(self, script, *args):
        """Execute script.

        :param script: Script to be executed.
        :param *args: Other arguments.

        :returns: Returns script result or None on failure.
        """
        try:
            return self.handle.execute_script(script, *args)
        except Exception as ex:
            self.exception(ex)
            return None

    def get_sizes(self):
        """Get sizes of the screen, window and content.

        :returns: Returns the sizes of current status, or None on failure.
        """
        try:
            sizes = self.execute("""
                return {
                    "screen.width": window.screen.width,
                    "screen.height": window.screen.height,
                    "screen.client.width": window.screen.availWidth,
                    "screen.client.height": window.screen.availHeight,
                    "window.width": window.outerWidth,
                    "window.height": window.outerHeight,
                    "window.client.width": window.innerWidth,
                    "window.client.height": window.innerHeight,
                    "page.width": document.documentElement.scrollWidth,
                    "page.height": document.documentElement.scrollHeight,
                    "page.client.width": document.documentElement.clientWidth,
                    "page.client.height": document.documentElement.clientHeight
                }
            """)
            if sizes is None: return None
            sizes["window.border.width"] = sizes["window.width"] - sizes["window.client.width"]
            sizes["window.border.height"] = sizes["window.height"] - sizes["window.client.height"]
            sizes["window.scrollbar.width"] = sizes["window.client.width"] - sizes["page.client.width"]
            sizes["window.scrollbar.height"] = sizes["window.client.height"] - sizes["page.client.height"]
            return sizes
        except:
            return None

    def get_element_path(self, element):
        """Get element's full path.

        :returns: Returns the full path of the element, or None on failure.
        """
        s = ""
        s += 'function metatutu_get_path(node){'
        s += '  if (node == null) return "";'
        s += '  if (node.parentNode == null) return "/.";'
        s += '  return metatutu_get_path(node.parentNode) + "/" + node.tagName;'
        s += '}'
        s += 'return metatutu_get_path(arguments[0]);'
        return self.execute(s, element)

    def save_screenshot(self, filepath):
        """Save screenshot as png file.

        :param filepath: File path.
        :returns: Returns True on success or False on failure.
        """
        return self.handle.save_screenshot(filepath)

    def get_screenshot_as_png(self):
        """Get screenshot as PNG data."""
        return self.handle.get_screenshot_as_png()

    def get_screenshot_as_base64(self):
        """Get screenshot as base64 data."""
        return self.handle.get_screenshot_as_base64()

    def get_screenshot_as_image(self):
        """Get screenshot as Image object.

        :returns: Returns Image object on success or None on failure.
        """
        try:
            temp_file = TempFile()
            if not self.save_screenshot(temp_file.path): return None
            image = ImageUtils.OpenImage(temp_file.path)
            return image
        except:
            return None

    def save_html(self, filepath):
        """Save current HTML as file.

        :param filepath: File path.
        :returns: Returns True on success or False on failure.
        """
        html = self.get_html()
        if html is None: return False
        return FileSystemUtils.save_file_contents(filepath, html)

    def capture_snapshot(self, filename, folderpath=None, with_html=True, with_screenshot=True):
        """Capture a snapshot.

        :param filename: Filename of the snapshot.  No extension is needed.
        :param folderpath: Where the file will be created.  If's None, use current working folder.
        :param with_html: Whether to save HTML.
        :param with_screenshot: Whether to save screenshot.
        """
        if folderpath is None: folderpath = ""
        filepath = os.path.join(os.path.abspath(folderpath), filename)
        if with_html: self.save_html(FileSystemUtils.alter_ext(filepath, ".html"))
        if with_screenshot: self.save_screenshot(FileSystemUtils.alter_ext(filepath, ".png"))



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import re
from urllib.parse import urlparse
from metatutu.logging import LoggerHelper
from metatutu.dc.scraping import HttpSession

class SiteScanner(LoggerHelper):
	"""Site scanner."""
	def __init__(self):
		LoggerHelper.__init__(self)

		self.search_paths = []
		self.detectors = {}

	def clear_search_paths(self):
		"""Remove all search paths."""
		self.search_paths = []

	def add_search_path(self, path):
		"""Add a search path.

		Search paths are alternative paths to be scanned in addition to the
		entrance URL.

		:param path: Path to be scanned.  eg. "/about"
		"""
		if path not in self.search_paths: self.search_paths.append(path)

	def clear_detectors(self):
		"""Remove all detectors."""
		self.detectors = {}

	def add_detector(self, name, patterns, exclusions):
		"""Add a detector.

		:param name: Detector name.  It will be used for output dataset as well.
		:param patterns: List of regex patterns to search the data from content.
		:param exclusions: List of regex patterns to exclude the data from search results.
			If the found data from content is fully matching any pattern in exclusions,
			it will not be included in the search results.
		"""
		self.detectors[name] = {
			"patterns": patterns,
			"exclusions": exclusions
		}

	def scan_content(self, content):
		"""Scan content and extract data.

		:param content: Content to be scanned.
		:returns: Returns a dict with found data of each detector or None on failure.
		"""
		try:
			#initialize data
			data = {}

			#go thru all detectors to extract data
			for name, detector in self.detectors.items():
				#initialize search results
				search_results = []

				#try each pattern to find the matching data
				for pattern in detector["patterns"]:
					candidates = re.findall(pattern, content)
					for candidate in candidates:
						#whether it had already been included
						if candidate in search_results: continue

						#whether it should be excluded
						excluded = False
						for exclusion in detector["exclusions"]:
							if re.fullmatch(exclusion, candidate):
								excluded = True
								break
						if excluded: continue

						#add to results
						search_results.append(candidate)

				#save search results
				data[name] = search_results

			#
			return data
		except Exception as ex:
			self.exception(ex)
			return None

	def get_page_content(self, url):
		"""Get content of the page at given URL.

		Derived classes could override this to implement different method
		to get the page content.

		:param url: URL of the page.
		:returns: Returns to content for scanning or None on failure.
		"""
		try:
			session = HttpSession()
			html = session.get_page(url, format="html", timeout=10.0)
			session.close()
			return html
		except Exception as ex:
			self.exception(ex)
			return None

	def scan_page(self, url):
		"""Scan page content and extract data.

		:param url: Page URL.
		:returns: Returns a dict with found data or None on failure.
		"""
		try:
			content = self.get_page_content(url)
			if content is None: return None
			return self.scan_content(content)
		except Exception as ex:
			self.exception(ex)
			return None

	def _get_search_urls(self, url, scan_search_paths_for_home_only):
		"""Get search URLs.

		Derived class could override it to add more search URLs based on
		searching strategies.  For example, to look into all linked pages
		on entrance page.

		:param url: Entrance page URL.
		:param scan_search_paths_for_home_only: see `scan_site()`.
		:returns: Returns a list of search URLs.
		"""
		search_urls = []
		urlparts = urlparse(url)
		if urlparts.scheme == "": return search_urls
		if urlparts.netloc == "": return search_urls
		if scan_search_paths_for_home_only:
			if urlparts.path not in ("", "/"): return search_urls
		base_url = urlparts.scheme + "://" + urlparts.netloc
		for search_path in self.search_paths:
			search_urls.append(base_url + search_path)
		return search_urls

	def _check_data_status(self, data):
		"""Check data status to see whether it should stop scanning.

		Derived class could override it to check the status of data, and judge
		whether the scanning could be stopped based on scanning strategy.
		The check will happen when switching page.

		:param data: Snapshot of data.
		:returns: Returns True to continue scanning.  It it returns False,
			scanning will be stopped.
		"""
		return True

	def scan_site(self, url, scan_search_paths_for_home_only=True):
		"""Scan site content and extract data.

		:param url: Entrance page URL.
		:param scan_search_paths_for_home_only: Whether to scan search paths if url is not for home page.
			If it's True, if url is not a home page, it will NOT scan search paths.
		:returns: Returns a dict with all data found in list by each detector regardless whether it's
			empty or whether any error had occurred.
		"""
		#initialize data
		data = {}
		for name in self.detectors.keys(): data[name] = []

		#get data
		try:
			#scan entrance page
			data_page = self.scan_page(url)
			if data_page:
				for name in self.detectors.keys():
					for item in data_page[name]:
						if item not in data[name]: data[name].append(item)

			#scan search paths
			search_urls = self._get_search_urls(url, scan_search_paths_for_home_only)
			for search_url in search_urls:
				if not self._check_data_status(data): break
				data_page = self.scan_page(search_url)
				if data_page:
					for name in self.detectors.keys():
						for item in data_page[name]:
							if item not in data[name]: data[name].append(item)
		except Exception as ex:
			self.exception(ex)

		#
		return data



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import time
import uuid
import pandas as pd
from IPython.display import *

__all__ = ["Display", "HTMLDisplay", "ProgressDisplay", "WorkflowWithProgressDisplay"]

class Display:
	"""Base class of Display classes.

	:ivar min_interval: Minimum interval between 2 refreshes.  In seconds.
	"""

	def __init__(self):
		# config
		self.min_interval = 0.5  # minimum interval between 2 refreshes

		# control
		self._display_id = None
		self._last_refresh_ts = time.time()

	def _refresh(self, content, force=False):
		if (not force) and (self._display_id is not None):
			if (time.time() - self._last_refresh_ts) < self.min_interval:
				return
		if self._display_id is None:
			self._display_id = str(uuid.uuid4())
			display(content, display_id=self._display_id)
		else:
			update_display(content, display_id=self._display_id)
		self._last_refresh_ts = time.time()

	def _display(self, content, force=False):
		self._refresh(content, force)

	def clear(self):
		self._refresh(HTML(""))

class HTMLDisplay(Display):
	"""Display HTML data in Jupyter Notebooks.

	This is also the base class of all other Display classes to be displayed
	with HTML content.
	"""

	def __init__(self):
		Display.__init__(self)

	def _get_html(self):
		return ""

	def _display_html(self, force=False):
		self._display(HTML(self._get_html()), force)

	@classmethod
	def display(cls, html):
		"""Display HTML content.

		:param html: HTML content.
		"""
		Display()._display(HTML(html), True)

	@classmethod
	def display_div(cls, html, width=None, height=None, style=None):
		"""Display HTML content in a <div> block.

		:param html: HTML content.
		:param width: Width of <div> block.  Optional.
		:param height: Height of <div> block.  Optional.
		:param style: Style of <div> block.  Optional.

		.. warning::
			When ``style`` is given, ``width`` and ``height`` will be ignored.
		"""
		div_style = ""
		if style:
			div_style = "style='{}'".format(style)
		else:
			style_width = ""
			if width is None:
				pass
			elif type(width) == str:
				style_width = "width: {}".format(width)
			elif type(width) == int:
				style_width = "width: {}px".format(width)

			style_height = ""
			if height is None:
				pass
			elif type(height) == str:
				style_height = "height: {}".format(height)
			elif type(height) == int:
				style_height = "height: {}px".format(height)

			if style_width != "" or style_height != "":
				div_style = "style='{}; {}'".format(style_width, style_height)

		cls.display("<div {}>{}</div>".format(div_style, html))

	@classmethod
	def display_head(cls, text, level=5):
		"""Display content in <h> tags.

		:param text: Text to be displayed.
		:param level: Level of head.
			For example, if level is 5, then ``text`` will be displayed in
			``<h5>`` tags.
		"""
		cls.display("<h{}>{}</h{}>".format(level, text, level))

	@classmethod
	def display_dataframe(cls, df, rows=None, width="100%", height=200, style=None):
		"""Display DataFrame object (pandas) in HTML format.

		:param df: DataFrame object.
		:param rows: Rows to be displayed.
			If it is None, then display all rows.
			If it is a positive number, then display number of head rows.
			If it is a negative number, then display number of tail rows.
		:param width: See ``display_div()``.
		:param height: See ``display_div()``.
		:param style: See ``display_div()``.

		The table heads (columns head) will be sticky.
		"""
		if type(df) != pd.DataFrame: return
		html_df = None
		if rows:
			if rows >= 0:
				html_df = df.head(rows).style.set_sticky(1).to_html()
			else:
				html_df = df.tail(-rows).style.set_sticky(1).to_html()
		else:
			html_df = df.style.set_sticky(1).to_html()
		cls.display_div(html_df, width, height, style)

class ProgressDisplay(HTMLDisplay):
	"""Progress display.

	:ivar show_text: Whether to show progress text.
	:ivar text_prefix: Text prefix.
	:ivar show_progress_bar: Whether to show progress bar.
	:ivar progress_bar_width: Progress bar width.
	:ivar hide_progress_bar_on_finished: Whether to hide the progress bar
		when workflow is finisihed.

	A typical usage for a workflow could be as below::

		#create progress display object
		p = ProgressDisplay()

		#before workflow start
		p.on_init(total_task_count)

		#workflow
		for task in tasks:
		    #update progress
		    p.on_update(completed_task_count)

		#after workflow
		p.on_finish()

	If you want to customize the progress text, create derived class and
	override ``get_text()`` to specify the text format.
	"""

	def __init__(self):
		HTMLDisplay.__init__(self)
		# config
		self.show_text = True
		self.text_prefix = ""
		self.show_progress_bar = True
		self.progress_bar_width = "100%"
		self.hide_progress_bar_on_finished = True
		self.hide_text_on_finished = False

		# control
		self._progress = 0
		self._total = 100
		self._finished = False
		self._start_ts = time.time()
		self._end_ts = self._start_ts

	def _get_html(self):
		html = "<div>"
		if self.show_progress_bar:
			if self._finished and self.hide_progress_bar_on_finished:
				pass
			else:
				html += "<progress style='width:{}' max='{}' value='{}'></progress>".format(
					self.progress_bar_width, self._total, self._progress)
		if self.show_text:
			if self._finished and self.hide_text_on_finished:
				pass
			else:
				html += "<p>{}</p>".format(self.get_text(self._finished))
		html += "</div>"
		return html

	def get_text(self, finished):
		"""Create progress text.

		This is called by framework to create progress text.
		Derived class may override it to customize the progress text.
		"""
		if self._total > 0:
			return "{} {} of {}/{:.1f}% (elapsed: {:.1f} seconds)".format(
				self.text_prefix,
				self._progress, self._total,
				self._progress / self._total * 100.0,
				self.elapsed())
		else:
			return "{} {} (elapsed: {:.1f} seconds)".format(
				self.text_prefix, self._progress, self.elapsed())

	def elapsed(self):
		"""Get time elapsed for the workflow.

		:returns: Time eplased since ``on_init()`` invoked.  In seconds.
		"""
		return self._end_ts - self._start_ts

	def on_init(self, total):
		"""Initialize progress.

		Call it before workflow is started.

		:param total: Total workload.
		"""
		self._progress = 0
		self._total = total
		self._finished = False
		self._start_ts = time.time()
		self._end_ts = self._start_ts
		self._display_html(True)

	def on_update(self, progress):
		"""Update progress.

		Call it during workflow to specify how much had been achieved.

		:param progress: Progress so far.
		"""
		self._end_ts = time.time()
		self._progress = progress
		self._display_html(False)

	def on_finish(self, total=None):
		"""Finish progress.

		Call it after workflow is finished.

		:param total: Reset the total workload for the cases total was
			unknown until finished.
		"""
		self._end_ts = time.time()
		if total: self._total = total
		self._finished = True
		self._display_html(True)

class WorkflowWithProgressDisplay:
	"""Simple workflow base class with progress display framework.

	.. note::
		Although it's recommended to use ProgressDisplay to show the progress,
		it could also use other alternative implementation.  This class is
		basically to provide a simple structure could be reused.
	"""

	def __init__(self, progress_display=None):
		self._p = progress_display

	def on_init(self, progress, total):
		if self._p: self._p.on_init(total)

	def on_update(self, progress, total):
		if self._p: self._p.on_update(progress)

	def on_finish(self, progress, total):
		if self._p: self._p.on_finish(total)

	def run(self):
		return False



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

from abc import ABC, abstractmethod
import threading
import queue
import time
import uuid

__all__ = [
	"Worker",
	"Workers",
	"TaskQueue",
	"Doer",
	"Operator",	"Controller", "Team",
	"Pipeline"]

class Worker(ABC, threading.Thread):
	"""Base class of worker classes.

	Worker is a processor run in its own working thread to play the specific
	role in a program in paraellel with other workers or program (main thread).

	This base class defines the framework of worker classes, so that the
	worker management will be standardized and flexible for different scale of
	pipeline.

	Program (main thread) takes below actions to manage a worker:

	1. Create an instance of worker class
	2. Call ``hire()`` to activate the worker
	3. Call ``dismiss()`` to deactivate the worker
	4. Delete the instance explicitly or have GC to collect it automatically

	Below methods are overrideable in derived worker class to meet the needs:

	* ``_process()``: Derived class must need to override it to implement the\
		process logic of a worker.  During the process, use ``self._dismisssNotice.is_set()``\
		to check whether program had sent notice of dismiss, and make sure\
		the exit of ``_process()`` as soon as possible to complete the dismiss\
		workflow.

	* ``_pre_working()``: Framework will call it right before the working\
		thread is started.  It's running in main thread.

	* ``_start_working()``: Framework will call it right after the working\
		thread is started.  It's running in working thread.

	* ``_stop_working()``: Framework will call it right before the working\
		thread is stopped.  It's running in working thread.

	* ``_post_working()``: Framework will call it right after the working thread\
		is stopped.  It's running in main thread.

	* ``_enter_idle()``: Framework or some process will call it when the worker\
		become idle(no task available) in task driven model.  It's running in\
		working thread.

	* ``_leave_idle()``: Framework or some process will call it when the worker\
		become busy(has task available) in task driven model.  It's running in\
		working thread.

	* ``bind()``: Framework will call it to bind the data passed to the framework.
	"""
	def __init__(self):
		threading.Thread.__init__(self)

		# control
		self.__stateLock = threading.Lock()
		self.__state = 0	# s0(init), s1(hired), s2(working), s3(dismissed)

		self._dismissNotice = threading.Event()

		self._idleLock = threading.Lock()
		self._idle = True

		# id
		self.id = ""

	def __del__(self):
		if self.__state == 0:
			pass
		elif self.__state == 1:
			raise Exception("dismiss() must be called explicitly!")
		elif self.__state == 2:
			raise Exception("dismiss() must be called explicitly!")
		else:
			pass

	@abstractmethod
	def _process(self): pass

	def _pre_working(self): pass
	def _start_working(self): pass
	def _stop_working(self): pass
	def _post_working(self): pass

	def _enter_idle(self):
		with self._idleLock: self._idle = True

	def _leave_idle(self):
		with self._idleLock: self._idle = False

	def _handle_exception(self, e): pass

	def bind(self, data): pass

	def hire(self, id=None):
		"""Hire a worker.

		:param id: Worker ID.  If it is None, an ID will be generated.

		:returns: Returns True on success and False on failure.
		"""
		# update state: s0(init)->s1(hired)
		with self.__stateLock:
			if self.__state != 0: return False
			self.__state = 1

		# set id
		if id is not None:
			self.id = id
		else:
			self.id = str(uuid.uuid4())

		# start working thread
		self._pre_working()
		try:
			self.start()
		except:
			return False

		#
		return True

	def dismiss(self):
		"""Dismiss the worker."""
		# check state
		with self.__stateLock:
			state = self.__state
			# s1(hired): working thread was not started during hire()
			# s2(working): normal cases
			if state != 1 and state != 2: return

		# stop working thread
		if state == 2:
			self._dismissNotice.set()
			try:
				self.join()
			except:
				pass
		self._post_working()

		# check state
		with self.__stateLock:
			if self.__state != 3: raise Exception("Unknown error.")

	def run(self):
		# update state: s1(hired)->s2(working)
		with self.__stateLock:
			self.__state = 2

		# process
		try:
			self._start_working()
			self._process()
			self._stop_working()
		except Exception as e:
			self._handle_exception(e)

		# update state: s2(working)->s3(dismissed)
		with self.__stateLock:
			self.__state = 3

	def is_idle(self):
		with self._idleLock:
			return self._idle

class Workers:
	"""Worker list."""
	def __init__(self):
		self.__listLock = threading.Lock()
		self.__list = []
		self.__total_count = 0
		self.__peak_count = 0

	def __del__(self):
		if len(self.__list) > 0:
			raise Exception("Not all workers had been dismissed!")

	def get_count(self):
		"""Get number of current workers.

		:returns: Number of current workers.
		"""
		with self.__listLock:
			return len(self.__list)

	def get_status(self):
		"""Get status of workers.

		:returns: Returns a dict with summary of workers status.
		"""
		with self.__listLock:
			total_count = self.__total_count
			peak_count = self.__peak_count
			count = 0
			idle_count = 0
			for worker in self.__list:
				count += 1
				if worker.is_idle(): idle_count += 1
		idle_rate = 0.0
		if count > 0: idle_rate = idle_count / count
		return {
			"total_count": total_count,		# number of workers had ever been hired
			"peak_count": peak_count,		# number of current workers at peak time
			"count": count,					# number of current workers
			"idle_count": idle_count,		# number of idle workers
			"idle_rate": idle_rate			# idle% = idle_count/count, 0 if no worker is available
		}

	def hire(self, workerClass, data=None, id=None):
		"""Hire a worker.

		:param workerClass: Worker class.
		:param data: Data to be passed to ``worker.bind()``.
		:param id: ID of the worker.  If it's None, it will generate an ID automatically.

		:returns: Number of workers had been hired.
		"""
		# hire a worker
		worker = workerClass()
		worker.bind(data)
		if not worker.hire(id): return 0

		# add worker to list
		with self.__listLock:
			self.__list.append(worker)
			self.__total_count += 1
			if len(self.__list) > self.__peak_count:
				self.__peak_count = len(self.__list)

		#
		return 1

	def hire_n(self, workerClass, n, data=None, ids=None):
		"""Hire n workers.

		:param workerClass: Worker class.
		:param n: Number of workers to be hired.
		:param data: Data to be passed to ``worker.bind()``.
		:param ids: List of IDs of the new workers.
			If it's None or not enough IDs in the list, id will be generated.

		:returns: Number of workers had been hired.
		"""
		if n <= 0: return 0
		count = 0
		for i in range(0, n):
			id = None
			if ids is not None:
				if i < len(ids): id = ids[i]
			if self.hire(workerClass, data, id) == 0: break
			count += 1
		return count

	def dismiss(self, id=None):
		"""Dismiss a worker.

		:param id: ID of the worker to be dismissed.
			If it's None, it will dismiss the last idle worker hired or\
			last worker hired.

		:returns: Number of workers had been dismissed.
		"""
		# remove worker from list
		with self.__listLock:
			count = len(self.__list)
			if count <= 0: return 0
			if id is not None:
				index = None
				for i in range(0, count):
					if self.__list[i].id == id:
						index = i
						break
				if index is None: return 0
			else:
				index = -1
				for i in range(0, count):
					if self.__list[count - i - 1].is_idle():
						index = count - i - 1
						break
			worker = self.__list.pop(index)

		# dismiss the worker
		worker.dismiss()
		del worker

		#
		return 1

	def dismiss_n(self, n):
		"""Dismiss n workers.

		:param n: Number of workers to be dismissed.

		:returns: Number of workers had been dismissed.
		"""
		if n <= 0: return 0
		count = 0
		for i in range(0, n):
			if self.dismiss() == 0: break
			count += 1
		return count

	def dismiss_all(self):
		"""Dismiss all workers.

		:returns: Number of workers had been dismissed.
		"""
		count = 0
		with self.__listLock:
			while len(self.__list) > 0:
				# remove worker from list
				worker = self.__list.pop(-1)

				# dismiss the worker
				worker.dismiss()
				del worker

				#
				count += 1

		#
		return count

class _TaskQueueItem(object):
	def __init__(self, index, priority, task):
		self.index = index
		self.priority = priority
		self.task = task

	def __lt__(self, other):
		if self.priority < other.priority:
			return True
		elif self.priority == other.priority:
			return self.index < other.index
		else:
			return False

class TaskQueue(queue.PriorityQueue):
	"""Task queue."""
	def __init__(self, maxsize=0):
		queue.PriorityQueue.__init__(self, maxsize)
		self.__total_count = 0
		self.__peak_count = 0

	def __del__(self):
		if not self.empty():
			pass

	def get_status(self):
		"""Get status of the queue.

		:returns: Returns a dict with summary of queue status.
		"""
		with self.mutex:
			total_count = self.__total_count
			peak_count = self.__peak_count
			count = self._qsize()
		return {
			"total_count": total_count,		# number of tasks had ever been queued
			"peak_count": peak_count,		# number of tasks at peak time
			"count": count					# number of current tasks
		}

	def push_task(self, task, priority=100, timeout=None):
		"""Push task to the queue.

		:param task: Task.
		:param priority: Priority of the task.  Lower value takes higher priority.
			Default value is 100.
		:param timeout: Timeout in seconds.  If it's None, it's infinite.

		:returns: Returns True on success and False on failure.
		"""
		try:
			if task is None: return False

			# get index
			with self.mutex:
				index = self.__total_count + 1

			# put
			self.put(_TaskQueueItem(index, priority, task), block=True, timeout=timeout)

			# update total & peak count
			with self.mutex:
				self.__total_count += 1
				if self._qsize() > self.__peak_count: self.__peak_count = self._qsize()
		except:
			return False
		return True

	def pop_task(self, timeout=None):
		"""Pop a task from the queue.

		:param timeout: If it's None, it will not wait for task if no task
			is available at the time.  If it's a postive value (in seconds),
			it will wait for task until time is out.

		:returns: Returns a task.  If no task is available, returns None.
		"""
		try:
			if timeout is None:
				taskItem = self.get(block=False)
			else:
				taskItem = self.get(block=True, timeout=timeout)
			task = taskItem.task
		except:
			task = None
		return task

class Doer(Worker):
	"""Base class of doer classes.

	Doer is the worker who is task drivern and typically working independently.
	It is with it's own task queue and be reponsible to finish all tasks
	in the queue before it's dismissed.

	Program pushes the tasks in the queue and just wait them processed by the doer.

	It defines a more specific framework and workflow than worker class.

	Override below methods to implement the logic for specific Doer:

	* ``_process_task()``: Called by framework when need to process a task.
	"""
	def __init__(self, maxsize=0):
		Worker.__init__(self)

		# task queue
		self.task_queue = TaskQueue(maxsize)

	def __del__(self):
		if self.task_queue is not None:
			del self.task_queue
			self.task_queue = None
		Worker.__del__(self)

	@abstractmethod
	def _process_task(self, task): pass

	def _process(self):
		stopping = False
		while True:
			# flush all tasks in queue as soon as possible
			self._leave_idle()
			while True:
				# pop a task
				task = self.task_queue.pop_task()
				if task is None: break

				# process the task
				self._process_task(task)
			self._enter_idle()

			# check stop flag
			if stopping: break

			# check stop request
			if self._dismissNotice.is_set():
				time.sleep(1)	# give one more chance to flush task queue
				stopping = True

class Operator(Worker):
	"""Base class of operator classes.

	Operator is the worker who is task drivern and working as part of team.
	It is using the shared task queue and be reponsible to finish current task
	only.

	Override below methods to implement the logic for specific Operator:

	* ``_pop_task()``: Called by framework when need to get a task from some task queue.

	* ``_process_task()``: Called by framework when need to process a task.
	"""
	def __init__(self):
		Worker.__init__(self)

	def __del__(self):
		Worker.__del__(self)

	@abstractmethod
	def _pop_task(self): return None

	@abstractmethod
	def _process_task(task): pass

	def _process(self):
		while True:
			# pop a task
			task = self._pop_task()
			if task is not None:
				# process the task
				self._leave_idle()
				self._process_task(task)
				self._enter_idle()

			# check stop request
			if self._dismissNotice.is_set(): break

class Controller(Worker):
	"""Base class of controller classes.

	Controller is the worker who is not task driven.
	It's typically working in parallel with other Operators to monitor and
	facilitate the pipeline.  Or it may take some special processes no matter
	it's once or repeating.
	"""
	def __init__(self):
		Worker.__init__(self)

	def __del__(self):
		Worker.__del__(self)

class Team(ABC):
	"""Base class of team classes.

	Team is an organization with Operators and Controllers to typically process
	a single kind of task.  Team is typically with:

	* ``task_queue``: A shared task queue.  Program could send the tasks to it,\
		and team is responsible to process the task in quality and timing.

	* ``operators``: A list of operators who is processing the tasks.  The list\
		could be dynamic based on the workload by controllers.

	* (controllers)`: Optional, to be defined in actual team class based on needs.

	Derived class needs to implement below logic as minimum:

	* ``__init__()``: Define the ``self.operator_class`` with operator class,\
		so that framework could manage operators automatically.

	* ``hire()``: Override it with logic to create the task queue, hire initial\
		operators and hire controllers.

	* ``dismiss()``: Override it with logic to wait for task queue empty, dismiss\
		operators, dismiss controllers, and clear and delete the task queue.
	"""
	def __init__(self):
		# task queue
		self.task_queue = None

		# operators
		self.operator_class = None
		self.operators = Workers()

	def __del__(self):
		if self.task_queue is not None:
			del self.task_queue
			self.task_queue = None

	@abstractmethod
	def hire(self): return False

	@abstractmethod
	def dismiss(self): pass

	def finish_all_tasks(self):
		while not self.task_queue.empty():
			if self.operators.get_count() <= 0:
				if self.hire_operator(1) < 1:
					raise Exception("Failed to hire operator to finish remaining tasks!")
			time.sleep(0.5)

	def hire_operator(self, count=1):
		if self.operator_class is None: return 0
		return self.operators.hire_n(self.operator_class, count, self, None)

	def dismiss_operator(self, count=1):
		return self.operators.dismiss_n(count)

class Pipeline(ABC):
	"""Base class of pipeline classes.

	Pipeline is an organization with Teams and Controllers to process complex
	tasks and jobs.  Pipeline is typically with:

	* (job queue): An input job queue.  Job could be considered as a complex\
		task which could be broken down into tasks.  Program could send the\
		job to the job queue.  Pipeline controller will break it down into\
		tasks based on the procedure and send the task to different Teams in\
		pipeline in order or in parallel.  Once all the tasks of a job had\
		been complete, the job could be delivered and leave the pipeline.

	* (teams): Teams for different kind of tasks.  Each team is focusing on\
		a single task type.  Teams could be working in parallel or in sequence\
		based on actual needs.

	* (controllers): Roles to facilitate the pipeline process.

	This class is not defining any workflow, and it's more link a model for
	developer to build the pipeline.  Override and implement below methods
	are suggested:

	* ``__init__()``: Define pipeline attributes.

	* ``hire()``: Logic to initialize the pipeline with create the job queue,\
		prepare resources, hire teams, hire controllers, etc.

	* ``dismiss()``: Logic to clean up the pipeline by handling remaining job\
		queue, dismiss teams, dismiss controllers, etc.
	"""
	def __init__(self):	pass

	def __del__(self): pass

	@abstractmethod
	def hire(self): return False

	@abstractmethod
	def dismiss(self): pass



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import os
import uuid
import datetime
import shutil
import re
import json
import tempfile

__all__ = ["FileSystemUtils", "TempFile", "TempFolder", "FileSystemDataStore"]

class FileSystemUtils:
	"""File system utilities."""

	@classmethod
	def create_folder(cls, folderpath):
		"""Create a folder.

		:param folderpath: Folder path.
		:returns: Returns True if folder exists, otherwise it returns False.
		"""
		try:
			if not os.path.exists(folderpath): os.makedirs(folderpath)
		except:
			pass
		return os.path.isdir(folderpath)

	@classmethod
	def create_parent_folder(cls, fpath):
		"""Create parent folder.

		:param fpath: File/folder path located under parent folder.
		:returns: Returns True if folder exists, otherwise it returns False.
		"""
		parent_folderpath = os.path.split(fpath)[0]
		return cls.create_folder(parent_folderpath)

	@classmethod
	def alter_fname(cls, ref_filepath, fname):
		"""Replace the file/folder name.

		:param ref_filepath: Reference file path.  eg. `__file__`
		:param fname: File/folder name.
		:returns: Returns the full path of new file/folder with new name.
		"""
		return os.path.join(os.path.dirname(os.path.abspath(ref_filepath)), fname)

	@classmethod
	def alter_ext(cls, ref_filepath, ext):
		"""Replace the extension.

		:param ref_filepath: Reference file path.  eg. `__file__`
		:param ext: File extension.  eg. ".log"
		:returns: Returns the full path of new file/folder with new extension.
		"""
		return os.path.splitext(os.path.abspath(ref_filepath))[0] + ext

	@classmethod
	def normalize_fname(cls, fname):
		"""Normalize filename.

		:param fname: Original filename.
		:returns: Returns the normalized filename.
		"""
		fname = fname.strip()
		c_list = ['\\', '/', ':', '*', '?', '<', '>', '|', '"', '\t', '\b', '\n', '\r']
		for c in c_list: fname = fname.replace(c, "")
		return fname

	@classmethod
	def file_exists(cls, filepath):
		"""Check whether file exists.

		:param filepath: File path.
		:returns: True if the file exists and False if it doesn't.
		"""
		return os.path.isfile(filepath)

	@classmethod
	def folder_exists(cls, folderpath):
		"""Check whether folder exists.

		:param folderpath: Folder path.
		:returns: True if the folder exists and False if it doesn't.
		"""
		return os.path.isdir(folderpath)

	@classmethod
	def delete_file(cls, filepath):
		"""Delete a file.

		:param filepath: File path.
		:returns: Result of action.
		"""
		try:
			os.remove(filepath)
		except:
			pass
		return not cls.file_exists(filepath)

	@classmethod
	def delete_folder(cls, folderpath):
		"""Delete a folder.

		:param folderpath: Folder path.
		:returns: Result of action.
		"""
		try:
			shutil.rmtree(folderpath)
		except:
			pass
		return not cls.folder_exists(folderpath)

	@classmethod
	def copy_file(cls, source_filepath, dest_filepath):
		"""Copy a file.

		:param source_filepath: Source file path.
		:param dest_filepath: Destination file path.
		:returns: Result of action.
		"""
		try:
			shutil.copy(source_filepath, dest_filepath)
		except:
			return False
		return True

	@classmethod
	def move_file(cls, source_filepath, dest_filepath):
		"""Move a file.

		:param source_filepath: Source file path.
		:param dest_filepath: Destination file path.
		:returns: Result of action.
		"""
		try:
			shutil.move(source_filepath, dest_filepath)
		except:
			return False
		return True

	@classmethod
	def list_all(cls, folderpath, recursive=False):
		"""List all files and folders in a folder.

		:param folderpath: Folder path.
		:param recursive: Whether to look into subfolders.
		:returns: Returns a list of all files/folders, or None on failure.
			List item is as (path, is_folder), `path` is the full path of the
			file/folder, and `is_folder` is boolean value to show whether it
			is a folder (True) or a file (False).
		"""
		def __walk(folderpath, l, recursive):
			fnames = os.listdir(folderpath)
			for fname in fnames:
				fpath = os.path.join(folderpath, fname)
				if os.path.isfile(fpath):
					l.append((fpath, False))
				else:
					l.append((fpath, True))
					if recursive:
						if not __walk(fpath, l, recursive): return False
			return True

		l = []
		if not __walk(os.path.abspath(folderpath), l, recursive): return None
		return l

	@classmethod
	def list(cls, folderpath, pattern=".*", recursive=False,
		match_fullpath=False,
		include_files=True, include_folders=True):
		"""List all files and folders matching pattern in a folder.

		:param folderpath: Folder path.
		:param pattern: RegEx pattern to match file/folder.  In lower cases.
		:param recursive:  Whether to look into subfolders.
		:param match_fullpath: Whether to match file/folder name or full path.
			If it's True, match full path.  Otherwise, match file/folder name.
		:param include_files: Whether to include files.
		:param include_folders: Whether to include folders.
		:returns: Returns a list of all filees/folders matching pattern, or None
			on failure.
		"""
		l = []
		items = cls.list_all(folderpath, recursive)
		if items is None: return None
		for fpath, is_folder in items:
			#file/folder filter
			if is_folder:
				if not include_folders: continue
			else:
				if not include_files: continue

			#pattern filter
			if match_fullpath:
				text = fpath.lower()
			else:
				text = os.path.basename(fpath).lower()
			if re.fullmatch(pattern, text): l.append((fpath, is_folder))
		return l

	@classmethod
	def list_files(cls, folderpath, pattern=".*", recursive=False, match_fullpath=False):
		"""List all files matching pattern in a folder.

		See `list()` for parameter definitions.

		:returns: Returns a list of file paths matching pattern, or None on failure.
		"""
		l = []
		items = cls.list(folderpath, pattern, recursive, match_fullpath, True, False)
		if items is None: return None
		for item in items: l.append(item[0])
		return l

	@classmethod
	def list_folders(cls, folderpath, pattern=".*", recursive=False, match_fullpath=False):
		"""List all folders matching pattern in a folder.

		See `list()` for parameter definitions.

		:returns: Returns a list of folder paths matching pattern, or None on failure.
		"""
		l = []
		items = cls.list(folderpath, pattern, recursive, match_fullpath, False, True)
		if items is None: return None
		for item in items: l.append(item[0])
		return l

	@classmethod
	def load_file_contents(cls, filepath, encoding="utf-8"):
		"""Load text file into a text string.

		:param filepath: File path.
		:param encoding: Encoding of text file.
		:returns: It returns a string with full text content.
			If it's failed, returns None.
		"""
		try:
			with open(filepath, "r", encoding=encoding) as f:
				contents = f.read()
		except:
			contents = None
		return contents

	@classmethod
	def save_file_contents(cls, filepath, contents, encoding="utf-8", create_parent=True):
		"""Save text string into a text file.

		:param filepath: File path.
		:param contents: Text content.
		:param encoding: Encoding of text file.
		:param create_parent: Whether to create parent before saving.
		:returns: Result of action.
		"""
		try:
			if create_parent: cls.create_parent_folder(filepath)
			with open(filepath, "w", encoding=encoding) as f:
				f.write(contents)
			return True
		except:
			return False

	@classmethod
	def load_file_bytes(cls, filepath):
		"""Load binary file into a bytes object.

		:param filepath: File path.
		:returns: It returns a bytes object with full binary content.
			If it's failed, returns None.
		"""
		try:
			with open(filepath, "rb") as f:
				contents = f.read()
		except:
			contents = None
		return contents

	@classmethod
	def save_file_bytes(cls, filepath, contents, create_parent=True):
		"""Save bytes object into a binary file.

		:param filepath: File path.
		:param contents: Bytes object.
		:param create_parent: Whether to create parent before saving.
		:returns: Result of action.
		"""
		try:
			if create_parent: cls.create_parent_folder(filepath)
			with open(filepath, "wb") as f:
				f.write(contents)
			return True
		except:
			return False

	@classmethod
	def load_file_json(cls, filepath, encoding="utf-8"):
		"""Load JSON file into a data object.

		:param filepath: File path.
		:returns: It returns an object as JSON data describes.
			If it's failed, returns None.
		"""
		try:
			with open(filepath, "r", encoding=encoding) as f:
				data = json.load(f)
			return data
		except:
			return None

	@classmethod
	def save_file_json(cls, filepath, data, encoding="utf-8", create_parent=True):
		"""Save a data object into a JSON file.

		:param filepath: File path.
		:param data: Data object.
		:param encoding: Encoding of JSON file.
		:param create_parent: Whether to create parent before saving.
		:returns: Result of action.
		"""
		try:
			if create_parent: cls.create_parent_folder(filepath)
			with open(filepath, "w", encoding=encoding) as f:
				json.dump(data, f, indent="\t")
			return True
		except:
			return False

class TempFileSystemObject:
	"""Base class of TempFile and TempFolder."""

	def __init__(self):
		self._is_folder = False
		self._fpath = None

	def __del__(self):
		self.delete()

	@property
	def path(self):
		"""Temp file/folder path."""
		return self._fpath

	def create(self, temp_folderpath=None):
		"""Create a temp file/folder.

		:param temp_folderpath: Temp folder path.

		:returns: Returns temp file/folder path if it's created successfully.
			Otherwise, it returns None.
		"""
		if temp_folderpath is None: temp_folderpath = tempfile.gettempdir()
		while True:
			fpath = os.path.join(temp_folderpath, str(uuid.uuid4()))
			if not os.path.exists(fpath): break
		if self._is_folder:
			if not FileSystemUtils.create_folder(fpath): return None
		else:
			FileSystemUtils.save_file_contents(fpath, "")
			if not FileSystemUtils.file_exists(fpath): return None
		return self.attach(fpath)

	def attach(self, fpath):
		"""Attach a file/folder as temp file.

		:param fpath: File/folder path.
		"""
		self.delete()
		self._fpath = fpath

	def detach(self):
		"""Detach the temp file/folder.

		.. warning::
			When the temp file/folder is detached, it will not be deleted
			automatically.  So there should be some logic to manage the
			file/folder to make sure they will not become garbage on file
			system.

		This is useful for download kind of use cases.  Before a file is
		fully downloaded, it is still in a temp status.  When it is fully
		downloaded, it could be detached and renamed to a permanent file.

		:returns: Detached temp file/folder path.
		"""
		fpath = self._fpath
		self._fpath = None
		return fpath

	def delete(self):
		"""Delete the temp file/folder."""
		if self._fpath:
			if self._is_folder:
				FileSystemUtils.delete_folder(self._fpath)
			else:
				FileSystemUtils.delete_file(self._fpath)
			self._fpath = None

class TempFile(TempFileSystemObject):
	"""Temp file."""

	def __init__(self, temp_folderpath=None):
		super().__init__()
		self._is_folder = False
		self.create(temp_folderpath)

class TempFolder(TempFileSystemObject):
	"""Temp folder."""

	def __init__(self, temp_folderpath=None):
		super().__init__()
		self._is_folder = True
		self.create(temp_folderpath)

class FileSystemDataStore(FileSystemUtils):
	r"""File system data store.

	This class is typically used to define a workspace on file system to store
	the application related data.

	It requires to specify a root folder as data store and all permanent or
	temporary data could be stored in this data store.

	It also enables the model of "dated" which could automatically managed the
	data by date.  For example, it could store the daily reports under "dated"
	folder with same filename.

	The class is derived from ``FileSystemUtils`` so that the usage of utilities
	could be easily called by a single object.

	Example::

		#global fsds object
		fsds = FileSystemDataStore(r"c:\data")

		#get a file path in workspace
		print(fsds.get_path("shared.dat"))
		# => c:\data\shared.dat

		#get a file path in workspace in date model
		print(fsds.get_dated_path("report.dat"))
		# => c:\data\2022-06-01\report.dat

		#call utilities with fsds object
		fsds.save_file_contents(fsds.get_path("raw.txt"), text)

	:ivar root_folderpath: Root folder path.
	:ivar temp_foldername: Temp folder name.
	"""

	def __init__(self, root_folderpath=None):
		#config
		self.root_folderpath = root_folderpath
		self.temp_foldername = ".temp"

	def get_path(self, fname, create_parent=False):
		"""Get full path of a file/folder in data store.

		:param fname: File/folder name.
		:param create_parent: Whether to create the parent folder.

		:returns: Full path of the file/folder.

		.. hint::
			``fname`` could be a specific name (eg. "abc.txt") or it could be
			with relative path included (eg. "reports/report1.txt").
		"""
		fpath = os.path.join(self.root_folderpath, fname)
		if create_parent: self.create_parent_folder(fpath)
		return fpath

	def get_parent_path(self, fname):
		"""Get the parent folder path of the file/folder in data store.

		:param fname: File/folder name.

		:returns: Full path of the parent folder path of the file/folder.
		"""
		return os.path.split(self.get_path(fname))[0]

	def get_dated_path(self, fname, date=None, create_parent=False):
		"""Get full path of a dated file/folder in data store.

		:param fname: File/folder name.
		:param date: Specify the date.  If it is None, use today's date.
		:param create_parent: Whether to create the parent folder.

		:returns: Full path of the dated file/folder.
		"""
		date_part = None
		if date:
			if type(date) == str:
				date_part = date
			elif type(date) == datetime.datetime:
				date_part = date.strftime("%Y-%m-%d")
		if date_part is None:
			date_part = datetime.datetime.today().strftime("%Y-%m-%d")
		fpath = os.path.join(self.root_folderpath, date_part, fname)
		if create_parent: self.create_parent_folder(fpath)
		return fpath

	def get_dated_parent_path(self, fname, date=None):
		"""Get the parent folder path of the dated file/folder in data store.

		:param fname: File/folder name.
		:param date: Specify the date.  If it is None, use today's date.

		:returns: Full path of the parent folder path of the dated file/folder.
		"""
		return os.path.split(self.get_dated_path(fname, date))[0]

	def create_temp_file(self):
		"""Create a temp file.

		:returns: Returns a TempFile object on success, otherwise it returns None.
		"""
		temp_file = TempFile(self.get_path(self.temp_foldername))
		if temp_file.path is None: return None
		return temp_file

	def create_temp_folder(self):
		"""Create a temp folder.

		:returns: Returns a TempFolder object on success, otherwise it returns None.
		"""
		temp_folder = TempFolder(self.get_path(self.temp_foldername))
		if temp_folder.path is None: return None
		return temp_folder



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import os
from setuptools import setup, find_packages

# get version information
fpath = os.path.abspath("./lib/metatutu/__version__.py")
about = {}
with open(fpath, "r", encoding="utf-8") as f:
    exec(f.read(), None, about)

# get detailed information
fpath = os.path.abspath("./README.md")
with open(fpath, "r", encoding="utf-8") as f:
    long_description = f.read()

# setup
setup(
    # version information
    name=about["__NAME__"],
    description=about["__DESC__"],
    version="{}.{}".format(about["__VERSION__"], about["__BUILD__"]),
    license=about["__LICENSE__"],
    author=about["__AUTHOR__"],
    author_email=about["__AUTHOR_EMAIL__"],
    url=about["__URL__"],
    project_urls=about["__PROJECT_URLS__"],

    # detailed information
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[],

    # package information
    package_dir={"": "lib"},
    packages=find_packages("lib"),
    include_package_data=True,

    # requirements
    platforms="any",
    python_requires="",
    setup_requires=[],
    install_requires=[]
)



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import numpy as np
from PIL import Image, ImageDraw, ImageFont
import imageio

class ImageUtils:
	"""Image utilities."""

	@classmethod
	def OpenImage(cls, filepath):
		"""Open and load the image.

		`PIL.Image.open` is a lazy load implementation, and it will not close
		the file until it's load.  Although it should be working with simply
		calling `Image.load()` right after `open(file_path)`, but according to
		the source code of PIL, the file will be still open if any error
		will be raised.  So to make sure the file is closed, this function is
		fully handling the file.

		According to the PIL documentation, this function may not work for
		image files with multiple frames.

		:param filepath: Image file path.
		:returns: Returns a `PIL.Image` object on success or None on failure.
		"""
		try:
			image = None
			with open(filepath, "rb") as fp:
				image = Image.open(fp)
				image.load()
			return image
		except:
			return None

class Images:
	"""Image list.

	It defines a list of Image objects and offer the tools for images processing.

	:ivar images: List of image objects.
	"""
	def __init__(self):
		self.images = []

	def clear(self):
		"""Clear image list."""
		self.images.clear()

	def append(self, image):
		"""Append an image to the list.

		:param image: Image object.
		"""
		self.images.append(image)

	def append_file(self, filepath, mode=None):
		"""Append an image from file to the list.

		:param filepath: Image file path.
		:param mode: Expected image mode in memory.
			If it is specified, it will convert the image into the mode.
			If it is None, then leave it as the image file format.
			See ``PIL.Image.MODES`` for details.

		:returns: Returns True on success or False on failure.
		"""
		try:
			if mode:
				image = Image.open(filepath).convert(mode)
			else:
				image = Image.open(filepath)
			self.images.append(image)
			return True
		except:
			return False

	def append_files(self, filepaths, mode=None):
		"""Append images from files to the list.

		:param filepaths: Image file path list.
		:param mode: See ``append_file()``.

		:returns: Returns True or success or False on failure.
		"""
		for filepath in filepaths:
			if not self.append_file(filepath, mode): return False
		return True

	def stitch(self, vertical=True, auto_crop=True):
		"""Stitch images into one image.

		:param vertical: If it's True, then all images will be stitched in
			**y** direction from up to down.  Otherwise, images will be
			stitched in **x** direction from left to right.
		:param auto_crop: Whether to crop the images to make sure they
			could be stitched together in specified direction.  If it's True,
			it will align the minimal width or height for vertical or horizontal
			direction.  If it's False, if images width or height are not aligned,
			it will return None to stop the process.

		:returns: Returns stitched image object on success or None on failure.
		"""
		# pass 1: get min size
		min_width = 1000000
		max_width = 0
		min_height = 1000000
		max_height = 0
		for image in self.images:
			width, height = image.size
			min_width = min(width, min_width)
			max_width = max(width, max_width)
			min_height = min(height, min_height)
			max_height = max(height, max_height)

		# pass 2： normalize size and create bitmaps
		bitmaps = []
		axis = 0
		if vertical:
			axis = 0
			if (not auto_crop) and (min_width != max_width): return None
			for image in self.images:
				width, height = image.size
				if width > min_width:
					bitmap = np.array(image.crop((0, 0, min_width, height)))
				else:
					bitmap = np.array(image)
				bitmaps.append(bitmap)
		else:
			axis = 1
			if (not auto_crop) and (min_height != max_height): return None
			for image in self.images:
				width, height = image.size
				if height > min_height:
					bitmap = np.array(image.crop((0, 0, width, min_height)))
				else:
					bitmap = np.array(image)
				bitmaps.append(bitmap)

		# stitch bitmaps and create stitched image
		return Image.fromarray(np.concatenate(tuple(bitmaps), axis=axis))

	def create_gif(self, filepath, duration, loop=0):
		"""Create GIF image file with images.

		:param filepath: Output GIF image file path.
		:param duration: Duration of each image frame.
		:param loop: Loop times.
		"""
		imageio.mimsave(filepath, self.images, "gif", duration=duration, loop=loop)



"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import os
import re
import psutil
import subprocess
from metatutu.fsds import FileSystemUtils, TempFile

class Cmdline:
    """Command line."""
    def __init__(self, cmdline=None):
        self._cmdline = None
        self._args = None
        if cmdline is None:
            self._args = []
        elif type(cmdline) is str:
            self._cmdline = cmdline
        elif type(cmdline) is list:
            self._args = cmdline
        elif type(cmdline) is Cmdline:
            self._cmdline = cmdline._cmdline
            self._args = cmdline._args

    @property
    def cmdline(self):
        if self._cmdline:
            return self._cmdline.strip()
        else:
            return self.args_to_cmdline(self._args, True).strip()

    def __str__(self):
        return self.cmdline

    def is_valid(self):
        return str(self) != ""

    def reset(self, arg0=None):
        self._cmdline = None
        self._args = []
        self.add(arg0)

    def add(self, arg):
        if arg: self._args.append(arg)

    @classmethod
    def args_to_cmdline(cls, args, auto_quoted=False):
        """Make command line text from argument list.

        This will make an argument list into a command line text (str) by
        connecting all arguments together.  It optionally adds the quotes
        automatically when `auto_quoted` is set to True to make the command
        working especially for those file paths with space in middle.

        For example, if the `args` is as below:
        ["notepad", "C:\\Users\\Test\\My Documents\\test.txt"]

        When `auto_quoted` is False, the created command line will be:
        * notepad C:\\Users\\Test\\My Documents\\test.txt

        When `auto_quoted` is True, the created command line will be:
        * notepad "C:\\Users\\Test\\My Documents\\test.txt"

        The rules to add quotes automatically are:
        * There is space
        * Not prefixed with "-" (eg. "-l 30", "--message=You got an error!")
        * Not prefixed with quotes (eg. "'C:\\Program Files'", '"D:\\Test Data"')

        :param args: List of arguments.
        :param auto_quoted: Whether to add quotes automatically.
        :returns: Command line text.
        """
        if auto_quoted:
            cmdline = ""
            for arg in args:
                add_quotes = False
                if arg.find(" ") >= 0:
                    if arg[0] not in ('-', '"', "'"):
                        add_quotes = True
                if cmdline != "": cmdline += " "
                if add_quotes:
                    cmdline += '"' + arg + '"'
                else:
                    cmdline += arg
        else:
            cmdline = " ".join(args)
        return cmdline

class OSUtils:
    """Operating system utilities."""

    @classmethod
    @property
    def is_windows(cls):
        return os.name == "nt"

    @classmethod
    @property
    def is_posix(cls):
        return os.name == "posix"

    @classmethod
    @property
    def pid(cls):
        return os.getpid()

    @classmethod
    def Process(cls, pid=None):
        """Create a process object with pid.

        :param pid: Process id.  If it's None, it means current process.
        :returns: Returns a process object or None on failure.
        """
        try:
            if pid is None: pid=cls.pid
            return psutil.Process(pid)
        except:
            return None

    @classmethod
    def process(cls, pid=None, p=None, cmdline=None):
        """Get a snapshot of process information.

        :param pid: Process id.  If both `pid` and `p` are None, it means current process.
            If `pid` is given, `p` will be ignored.
        :param p: Process object.  This will be used only when `pid` is None.
        :param cmdline: Command line text to be filled.  If it's None, it will be generated.
        :returns: Returns a dict with process information or None on failure.
        """
        try:
            if pid:
                p = cls.Process(pid)
                if p is None: return None
            else:
                if p:
                    pass
                else:
                    p = cls.Process()
                    if p is None: return None
            return {
                "pid": p.pid,
                "name": p.name(),
                "cmdline": cmdline if cmdline else Cmdline.args_to_cmdline(p.cmdline(), False),
                "create_time": p.create_time(),
                "cwd": p.cwd(),
                "username": p.username(),
                "ppid": p.ppid()
            }
        except:
            return None

    @classmethod
    def list_processes(cls, pattern=".*"):
        """Get a snapshot of all processes meeting search criteria.

        :param pattern: Pattern of command line for searching, in RegEx and lower cases.
        :returns: Returns a list of processes found.
            List is as [{process}].
        """
        processes = []
        pids = psutil.pids()
        for pid in pids:
            try:
                p = psutil.Process(pid)
                cmdline = Cmdline.args_to_cmdline(p.cmdline(), False)
                if not re.fullmatch(pattern, cmdline.lower()): continue
                process = cls.process(p=p, cmdline=cmdline)
                if process: processes.append(process)
            except:
                pass
        return processes

    @classmethod
    def list_subprocesses(cls, pid=None, pattern=".*"):
        """Get a snapshot of all subprocesses meeting search criteria.

        :param pid: Process id.  If it's None, it means current process.
        :param pattern: Pattern of command line for searching, in RegEx and lower cases.
        :returns: Returns a list of processes found, or None on failure.
            List is as [{process}].
        """
        processes = []
        try:
            pp = cls.Process(pid)
            if pp is None: return None
            for p in pp.children():
                cmdline = Cmdline.args_to_cmdline(p.cmdline(), False)
                if not re.fullmatch(pattern, cmdline.lower()): continue
                process = cls.process(p=p, cmdline=cmdline)
                if process: processes.append(process)
        except:
            pass
        return processes

    @classmethod
    def list_open_files(cls, pattern=".*"):
        """Get a snapshot of all opened files from processes meeting search criteria.

        :param pattern: Pattern of command line for searching, in RegEx and lower cases.
        :returns: Returns a dict of all open files with the list of processes found.
            Dict is as {filepath: [{process}]}.
        """
        open_files = {}
        pids = psutil.pids()
        for pid in pids:
            try:
                p = psutil.Process(pid)
                cmdline = Cmdline.args_to_cmdline(p.cmdline(), False)
                if not re.fullmatch(pattern, cmdline.lower()): continue
                pofs = p.open_files()
                for pof in pofs:
                    file_path = pof.path
                    if file_path in open_files.keys():
                        open_files[file_path].append(cls.process(p=p, cmdline=cmdline))
                    else:
                        open_files[file_path] = [cls.process(p=p, cmdline=cmdline)]
            except:
                pass
        return open_files

    @classmethod
    def who_open_the_file(cls, file_path, pattern=".*"):
        """Find processes who is opening the file.

        :param file_path: File path.
        :param pattern: Pattern of command line for searching, in RegEx and lower cases.
        :returns: Returns a list of processes who is opening the file.
            List is as [{process}].
        """
        fpath_target = os.path.abspath(file_path).lower()
        open_files = cls.list_open_files(pattern)
        processes = []
        for fpath, p_list in open_files.items():
            if not fpath.lower() == fpath_target: continue
            processes += p_list
        return processes

    @classmethod
    def create_subprocess(cls, cmdline, wait=False, mode="null", output=None, **kwargs):
        """Create a process.

        This is a super implementation based on `class subprocess.Popen`
        to create the process with predefined modes, so that it will be
        easier for application developers.

        Below modes are available:

        * "null": Subprocess, no output.

        * "new": Subprocess, output to the new console.  (Windows only)

        * "current": Subprocess, output to the current console.

        * "output": Subprocess, output to the `output['stdout']`.  This mode
            doesn't work when `wait` is not True.

        :param cmdline: Command line.  Could be str, list of arguments or
            an instance of `Cmdline`.
        :param wait: If it's True, it will wait for subprocess terminated.
            Otherwise, it will returns immediately after subprocess created.
        :param mode: Mode.  If it's None, it will bypass the special settings.
        :param output: Extra output values in dict.  If's None, ignore all
            extra output.
        :returns: Returns a `class subprocess.Popen` object on success,
            or None on failure.
        """
        tf_stdout = None
        f_stdout = None
        try:
            #get settings
            stdout = kwargs.pop("stdout", None)
            creationflags = kwargs.pop("creationflags", 0)

            #apply special settings based on mode
            if mode:
                if mode == "null":
                    stdout = subprocess.DEVNULL
                elif mode == "new":
                    if cls.is_windows:
                        stdout = None
                        creationflags = subprocess.CREATE_NEW_CONSOLE
                elif mode == "current":
                    stdout = None
                elif mode == "output":
                    if not wait: return None
                    tf_stdout = TempFile()
                else:
                    return None

            #output to file
            if tf_stdout:
                f_stdout = open(tf_stdout.path, "w")
                stdout = f_stdout

            #create subprcess
            p = subprocess.Popen(
                Cmdline(cmdline).cmdline,
                stdout=stdout,
                creationflags=creationflags,
                **kwargs)

            #wait
            if wait: p.wait()

            #get output
            if f_stdout:
                f_stdout.close()
                f_stdout = None
                content = FileSystemUtils.load_file_contents(tf_stdout.path)
                if content is None: content = ""
                if output is not None: output["stdout"] = content

            #
            return p
        except:
            return None
        finally:
            if f_stdout: f_stdout.close()
            if tf_stdout: tf_stdout.delete()

    @classmethod
    def run(cls, cmdline, run_as_command=False, **kwargs):
        """Run a command with shell or an application without shell.

        :param cmdline: Command line.
        :param run_as_command: Whether to run command line as command or application.
            True to run it as command with shell, False to run it as application
            without shell.
        :returns: Returns return code of the command.  Returns None on failure.
        """
        kwargs.pop("shell", None)
        kwargs.pop("wait", None)
        mode = kwargs.pop("mode", None)
        if run_as_command:
            if mode == "new":  #for Windows only
                p = cls.create_subprocess("cmd /c " + cmdline, wait=True, shell=False, mode=mode, **kwargs)
            else:
                p = cls.create_subprocess(cmdline, wait=True, shell=True, mode=mode, **kwargs)
        else:
            p = cls.create_subprocess(cmdline, wait=True, shell=False, mode=mode, **kwargs)
        if p is None: return None
        return p.returncode

    @classmethod
    def start(cls, cmdline, **kwargs):
        """Start an application without shell."""
        kwargs.pop("shell", None)
        kwargs.pop("wait", None)
        return cls.create_subprocess(cmdline, wait=False, shell=False, **kwargs)

    @classmethod
    def kill_subprocesses(cls):
        """Kill all subprocesses."""
        try:
            pp = cls.Process()
            if pp is None: return
            for p in pp.children(): p.kill()
        except:
            pass

    @classmethod
    @property
    def PATH(cls):
        try:
            return os.environ["PATH"].split(os.pathsep)
        except:
            return []

    @classmethod
    def where(cls, filename, extra_paths=None):
        """Find file in current folder, extra folders and folders in PATH.

        :param filename: File name.
        :param extra_paths: A list of folder paths to be searched.  Or None to skip.
        :returns: Returns a list of full path of found files in order."""
        filepaths = []
        filepath = os.path.join(os.getcwd(), filename)
        if os.path.isfile(filepath): filepaths.append(filepath)
        search_paths = []
        if extra_paths: search_paths += extra_paths
        search_paths += cls.PATH
        for path in search_paths:
            filepath = os.path.join(path, filename)
            if os.path.isfile(filepath):
                if filepath not in filepaths: filepaths.append(filepath)
        return filepaths


"""
    This is part of METATUTU library.
    https://pypi.org/project/metatutu/

	:author: max.wu@wooloostudio.com
	:copyright: Copyright (C) 2022 Wooloo Studio.  All rights reserved.
	:license: see LICENSE.
"""

import os
import time
import threading
import traceback
from abc import ABC, abstractmethod
import metatutu.pipeline

__all__ = [
	"MessageLevel",
	"Logger",
	"TextLogger",
	"ConsoleLogger",
	"TextFileLogger", "FileLogger", "DatedFileLogger",
	"Loggers",
	"LoggerHelper"]

class MessageLevel:
	"""Message level."""
	GENERAL = 5
	DEBUG = 10
	INFO = 20
	WARNING = 30
	ERROR = 40
	CRITICAL = 50
	SEPARATOR = 100

class Logger(ABC):
	"""Base class of logger classes.

	It's an abstract class, implementation of specific logger needs to override
	`_log()` to handle the logging message output in certain format to certain
	output device.

	:ivar enabled_level: Minimal level of message to be logged.
		It's kind of filter of logging messages.  All messages at `enabled_level`
		or higher will be logged.  Otherwise, it will be ignored.  For example,
		if it's set to `MessageLevel.CRITICAL`, then only messages at level of
		50 or above will be logged.
	:ivar async_mode: Whether to use async mode to log messages.
		Async mode is the way to avoid program be blocked by the logging.
		It will cache the logging message and the message will be logged to
		particular devices in a separate thread.
	:ivar async_queuesize: Size of logging message queue (cache).
		It' valid for async mode only to specify how many messages would be
		cached at peak time.
	:ivar async_timeout: Timeout of caching message.
		This is valid for async mode only.  When logging message queue (cache)
		is full, even under async mode, calling thread will be blocked.
		This value is to define how long it will wait or the message will be
		discarded.  Set it to None to wait infinitely.

	.. warning::
		When the logger is running under async mode, make sure `close()` will be
		called explicitly, otherwise, the main program could be dead locked due
		the threading handling mechanism in Python.
	"""
	def __init__(self):
		# config
		self.enabled_level = 0  # record all messages (level >= 0)
		self.async_mode = False
		self.async_queuesize = 100
		self.async_timeout = None

		# control
		self.__async_logger = None
		self._log_lock = threading.Lock()

	def __del__(self):
		if self.__async_logger:
			raise Exception("close() must be called explicitly for async mode.")

	class __AsyncLogger(metatutu.pipeline.Doer):
		def __init__(self, maxsize, logger):
			super().__init__(maxsize)
			self.logger = logger

		def _process_task(self, task):
			with self.logger._log_lock:
				self.logger._log(task[0], task[1], task[2], task[3])

	def open(self):
		"""Open the logger.

		No need to call it explicitly.  Framework will call it automatically.
		"""
		if self.async_mode:
			if self.__async_logger is None:
				self.__async_logger = self.__AsyncLogger(self.async_queuesize, self)
				self.__async_logger.hire()

	def close(self):
		"""Close the logger.

		It's required to call it explicitly for async mode.
		"""
		if self.__async_logger:
			self.__async_logger.dismiss()
			del self.__async_logger
			self.__async_logger = None

	@abstractmethod
	def _log(self, timestamp, message, depth, level): pass

	def __log(self, timestamp, message, depth, level, std_level):
		# open anyway
		self.open()

		# log
		if level >= self.enabled_level:
			if self.async_mode:
				self.__async_logger.task_queue.push_task(
					(timestamp, message, depth, std_level), 100, self.async_timeout
				)
			else:
				with self._log_lock:
					self._log(timestamp, message, depth, std_level)

	def log(self, message, depth=0, level=MessageLevel.GENERAL):
		self.__log(time.time(), message, depth, level, MessageLevel.GENERAL)

	def debug(self, message, depth=0, level=MessageLevel.DEBUG):
		self.__log(time.time(), message, depth, level, MessageLevel.DEBUG)

	def info(self, message, depth=0, level=MessageLevel.INFO):
		self.__log(time.time(), message, depth, level, MessageLevel.INFO)

	def warning(self, message, depth=0, level=MessageLevel.WARNING):
		self.__log(time.time(), message, depth, level, MessageLevel.WARNING)

	def error(self, message, depth=0, level=MessageLevel.ERROR):
		self.__log(time.time(), message, depth, level, MessageLevel.ERROR)

	def critical(self, message, depth=0, level=MessageLevel.CRITICAL):
		self.__log(time.time(), message, depth, level, MessageLevel.CRITICAL)

	def separator(self, ch="=", width=80, level=MessageLevel.SEPARATOR):
		self.__log(time.time(), ch * width, 0, level, MessageLevel.SEPARATOR)

	def exception(self, ex, depth=0, level=MessageLevel.DEBUG):
		message = "EXCEPTION: " + str(ex) + "\n"
		message += traceback.format_exc()
		self.__log(time.time(), message, depth, level, MessageLevel.GENERAL)

class TextLogger(Logger):
	"""Base class of text logger classes.

	This is the base class and defines the common behaviors of the loggers
	with text based output.

	:ivar include_timestamp: Whether to include the timestamp.
		This is useful to use logger for console applications.  For the
		logger on console, it could hide the timestamp to show the sequence
		with logging messages only.
	:ivar use_utc: Whether to use UTC or local time for timestamp.
	:ivar indent_size: Size of indent for the depth.
		Depth is the unique functionality in this logging system, to make
		the calling stack for complex workflow clearer.  For the logging
		messages for each depth, it will be displayed in hierachy format.
	:ivar vline: The charactor to be used to represent the vertical line.
	"""
	def __init__(self):
		Logger.__init__(self)

		# config
		self.include_timestamp = True
		self.use_utc = False
		self.indent_size = 4
		self.vline = "|"

	def format_output(self, timestamp, message, depth, level):
		# separator
		if level == MessageLevel.SEPARATOR: return message

		# create timestamp text
		ts_text = ""
		if self.include_timestamp:
			if self.use_utc:
				ts_text = time.strftime("%Y-%m-%d %H:%M:%SZ ", time.gmtime(timestamp))
			else:
				ts_text = time.strftime("%Y-%m-%d %H:%M:%S ", time.localtime(timestamp))

		# create message text
		message_text = ""
		if level == MessageLevel.DEBUG:
			message_text = message
		elif level == MessageLevel.INFO:
			message_text = "INFO: " + message
		elif level == MessageLevel.WARNING:
			message_text = "WARNING: " + message
		elif level == MessageLevel.ERROR:
			message_text = "ERROR: " + message
		elif level == MessageLevel.CRITICAL:
			message_text = "CRITICAL: " + message
		else:
			message_text = message
		message_lines = message_text.replace("\r\n", "\n").replace("\r", "\n").split("\n")
		if len(message_lines) == 0: return ts_text

		# create indent text
		indent_size = self.indent_size
		indent_text = ""
		if depth < 0: depth = 0
		if indent_size < 0: indent_size = 0
		if indent_size < 3:
			indent_text = " " * indent_size
		else:
			if len(self.vline) == 1:
				indent_text = self.vline + " " * (indent_size - 1)
			else:
				indent_text = " " * indent_size

		# format output
		output_text = ""
		for i in range(0, len(message_lines)):
			if i == 0:
				output_text = ts_text + indent_text * depth + message_lines[i]
			else:
				output_text += "\n" + " " * len(ts_text) + indent_text * depth + message_lines[i]

		#
		return output_text

class ConsoleLogger(TextLogger):
	"""Logger for console."""
	def __init__(self):
		TextLogger.__init__(self)

	def _log(self, timestamp, message, depth, level):
		output_text = self.format_output(timestamp, message, depth, level)
		print(output_text)

class TextFileLogger(TextLogger):
	"""Base class of logger classes for text file (log file).

	:ivar encoding: Encoding of text file.
	:ivar mode: Open mode of the text file.
	:ivar auto_flush: Whether to flush on each log activity.
	"""
	def __init__(self):
		TextLogger.__init__(self)

		# config
		self.encoding = "utf-8"
		self.mode = "a"
		self.auto_flush = True

		# control
		self._files = []

	def __del__(self):
		self.close()

	def close(self):
		TextLogger.close(self)

		# close all files
		for file in self._files:
			if file is not None: file.close()
		self._files.clear()

	@abstractmethod
	def _get_file(self, timestamp, message, depth, level): pass

	def _log(self, timestamp, message, depth, level):
		output_text = self.format_output(timestamp, message, depth, level)
		file = self._get_file(timestamp, message, depth, level)
		if file:
			print(output_text, file=file)
			if self.auto_flush: file.flush()

class FileLogger(TextFileLogger):
	"""Logger for a standalone log file."""
	def __init__(self, filepath, reset=False):
		TextFileLogger.__init__(self)

		# control
		self._files = [None]
		self._filepath = filepath
		if reset: self.mode = "w"

	def _get_file(self, timestamp, message, depth, level):
		f = self._files[0]
		if f: return f
		try:
			f = open(self._filepath, mode=self.mode, encoding=self.encoding)
			if f: self._files[0] = f
		except:
			f = None
		return f

class DatedFileLogger(TextFileLogger):
	"""Logger for dated log files."""
	def __init__(self, folderpath):
		TextFileLogger.__init__(self)

		# control
		self._files = [None]
		self._folderpath = folderpath
		self._current_filepath = ""

	def _get_file(self, timestamp, message, depth, level):
		# get file path
		filename = ""
		if self.use_utc:
			filename = time.strftime("%Y-%m-%d.log", time.gmtime(timestamp))
		else:
			filename = time.strftime("%Y-%m-%d.log", time.localtime(timestamp))
		filepath = os.path.join(self._folderpath, filename)
		if filepath == self._current_filepath: return self._files[0]

		# close current file
		if self._files[0]:
			self._files[0].close()
			self._files[0] = None
			self._current_filepath = ""

		# open file
		f = None
		try:
			f = open(filepath, mode=self.mode, encoding=self.encoding)
			if f:
				self._files[0] = f
				self._current_filepath = filepath
		except:
			f = None
		return f

class Loggers:
	"""Aggregator of loggers.

	It will distribute the log messages to all loggers registered.
	"""
	def __init__(self):
		self.loggers = []

	def __del__(self):
		self.close()

	def close(self):
		for logger in self.loggers: logger.close()
		self.loggers.clear()

	def register_logger(self, logger):
		"""Register a logger.

		:param logger: A logger instance.
		"""
		self.loggers.append(logger)

	def log(self, message, depth=0, level=MessageLevel.GENERAL):
		for logger in self.loggers: logger.log(message, depth, level)

	def debug(self, message, depth=0, level=MessageLevel.DEBUG):
		for logger in self.loggers: logger.debug(message, depth, level)

	def info(self, message, depth=0, level=MessageLevel.INFO):
		for logger in self.loggers: logger.info(message, depth, level)

	def warning(self, message, depth=0, level=MessageLevel.WARNING):
		for logger in self.loggers: logger.warning(message, depth, level)

	def error(self, message, depth=0, level=MessageLevel.ERROR):
		for logger in self.loggers: logger.error(message, depth, level)

	def critical(self, message, depth=0, level=MessageLevel.CRITICAL):
		for logger in self.loggers: logger.critical(message, depth, level)

	def separator(self, ch="=", width=80, level=MessageLevel.SEPARATOR):
		for logger in self.loggers: logger.separator(ch, width, level)

	def exception(self, ex, depth=0, level=MessageLevel.DEBUG):
		for logger in self.loggers: logger.exception(ex, depth, level)

class LoggerHelper:
	"""Helper for classes need logging.

	This is to be used to bind a logger with a class which needs logging
	functions.
	"""
	def __init__(self):
		self._logger = None

	@property
	def logger(self):
		return self._logger

	def bind_logger(self, logger):
		self._logger = logger

	def log(self, message, depth=0, level=MessageLevel.GENERAL):
		if self._logger: self._logger.log(message, depth, level)

	def debug(self, message, depth=0, level=MessageLevel.DEBUG):
		if self._logger: self._logger.debug(message, depth, level)

	def info(self, message, depth=0, level=MessageLevel.INFO):
		if self._logger: self._logger.info(message, depth, level)

	def warning(self, message, depth=0, level=MessageLevel.WARNING):
		if self._logger: self._logger.warning(message, depth, level)

	def error(self, message, depth=0, level=MessageLevel.ERROR):
		if self._logger: self._logger.error(message, depth, level)

	def critical(self, message, depth=0, level=MessageLevel.CRITICAL):
		if self._logger: self._logger.critical(message, depth, level)

	def separator(self, ch="=", width=80, level=MessageLevel.SEPARATOR):
		if self._logger: self._logger.separator(ch, width, level)

	def exception(self, ex, depth=0, level=MessageLevel.DEBUG):
		if self._logger: self._logger.exception(ex, depth, level)
